/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
/**********************************************************************
  drawing area
*********************************************************************/
package parka;

import java.awt.*;
import java.util.*;

public class DrawArea extends java.awt.Panel
{
    // size of virtual screen
    final static int MAX_SIZE = 1000;
    // background color
    final static Color BG_COLOR = Color.white;

    //default location for creation of new Node:
    final static int  DEFAULT_NEWNODE_XPOS = 30;
    final static int  DEFAULT_NEWNODE_YPOS = 30;

    private int       offset_x, offset_y;    //scrollbar values
    private Scrollbar hbar, vbar;

    private QueryWin parent;
    private ToolBar  tb;

    private Image    offscreenImg;
    private Graphics imageG;

    private  int  pendLinkX, pendLinkY;
                  //temp storage of mouse loc for pending link
    private  int  pendingXPos;     //temp mouse x loc for pending edge
    private  boolean  isPendingEdge;
    private  boolean  isPendingLink;
    private  int      mouseDragX, mouseDragY;
		// position of mouse during a pending Edge draw

    private Class    activeClass;
    private Class    DEFAULT_CLASS; //treat as const, after initialization

    // Font information
    private  Font        f;
    private  FontMetrics fm;

    // *******
    // Methods
    // *******

    public DrawArea (QueryWin MyWin, ToolBar tb)
    {
        //implicit super() call here creates the panel
        
        // scrolling setup
        hbar = create_scrollbar (Scrollbar.HORIZONTAL, 0, 0, 0, MAX_SIZE-1);
        vbar = create_scrollbar (Scrollbar.VERTICAL, 0, 0, 0, MAX_SIZE-1);
        offset_x = 0;
        offset_y = 0;

	initTempLinkData();
	initTempEdgeData();
        setClass ("java.lang.Object");
        DEFAULT_CLASS = activeClass;  //init var; treat as const after this

        // remember your window
        parent = MyWin;
        // and your associated toolbar
        this.tb = tb;

        // Font setup
        f = new Font(parent.fontname, Font.PLAIN, parent.fontsize);
        fm = parent.parent.getFontMetrics(f);

        // Layout
        setLayout (new BorderLayout());
        add ("South", hbar);
        add ("East", vbar);

        // get all events
        requestFocus();
    }

    Scrollbar create_scrollbar (int orient, int value, int vis,
                                     int min, int max)
    {
	Scrollbar  b = new Scrollbar (orient, value, vis, min, max);
	/* value = init value of sb; vis (visibility) = visible part of
           component "below"; min sb value; max sb value   */
	b.setLineIncrement (MAX_SIZE/100);  /* unit increment of sb */
	b.setPageIncrement (MAX_SIZE/10);  /* block/page increment of sb */

	return  b;
    }

    public void update (Graphics g)
    {
	//System.out.println ("DrawArea.update()");

	paint (g);
    }

    public void paint (Graphics g)
    {
	//System.out.println ("DrawArea.paint()");

	if (ApplType.getIsVersion1_1())
            super.paint(g);  /* Parka object tries to call Container.paint()
                 when inside netscape -- possibly because this is compiled
                 with a 1.1 compiler, so we avoid this.  Component.paint()
                 just calls this paint again, and would be an infinite loop
                 in netscape; so there is no else clause. */

        // erase previous off screen image
        imageG.setColor(BG_COLOR);
        imageG.fillRect(0,0,bounds().width, bounds().height);

	// draw to off screen image
        imageG.setColor(getForeground());
        drawGraphObjects (imageG);

	if (isPendingEdge)
	    showPendingEdge (mouseDragX, mouseDragY);

	// copy image to window
        g.drawImage (offscreenImg, 0, 0, this);
    }

    protected FontMetrics getDAFontMetrics()
    {
	return  fm;
    }

    protected Font getDAFont()
    {
	return  f;
    }

    // x scrollbar value
    //
    int getXSBoffset()
    {
	return  offset_x;
    }

    // y scrollbar value
    //
    int getYSBoffset()
    {
	return  offset_y;
    }

    GraphObject getActiveOb()
    {
	return  (GraphObject.getActiveOb());
    }

    // Draw in order given, so nodes clip lines from links when they
    // should be clipped (by nodes) and not when they shouldn't (by edges).
    //
    public void drawGraphObjects (Graphics g)
    {
	Link.drawLinks (g);
	Edge.drawEdges (g);
	Node.drawNodes (g);
    }

    // Delete all graph objects
    //
    void clearGraph()
    {
	Link.clearAll();
	Node.clearAll();
	Edge.clearAll();

	unselectObject();
    }

    void unselectObject()
    {
        GraphObject.unselectObject (this);    //does redraw
    }

    void translate()
    {
        offset_x = hbar.getValue();
        offset_y = vbar.getValue();
        repaint();
    }

    //when window is moved or resized
    public synchronized void reshape(int x, int y, int width, int height)
    {
        super.reshape(x, y, width, height);
   
        Dimension d = size();
        if (d.width == 0)     d.width=1;
        if (d.height == 0)    d.height=1;

	//System.out.println ("DrawArea.reshape:  createImage() called next.");

        offscreenImg = createImage (d.width, d.height);
	     // creates offscreen drawable image to use in double buffering
        imageG = offscreenImg.getGraphics();
	updateFont (f, fm);

        hbar.setValues (hbar.getValue(), width,hbar.getMinimum(),
                                hbar.getMaximum());
        vbar.setValues (vbar.getValue(),height, vbar.getMinimum(),
                                vbar.getMaximum());
	// paint will be called by this routine.  Expose events (in X terms)
	// seem to only generate 'paint()' calls directly; not calls to
	// 'update()'.  [X/Motif is the peer on the system this is tested on].
    }

    void askForPredicate()
    {
/* TEST
	System.out.println ("Entered 'askForPredicate()'");
*/
	String[]   predlist = parent.pn.getPredicateList();

	PredDialog p = new PredDialog (parent, this, predlist);

	if (ApplType.getIsVersion1_0())
	    ((Window)p).show();  //Netscape thinks we want Dialog.show()
                         //which doesn't exist until java 1.1
	else
	    p.show();
    }

    void initTempLinkData()
    {
	pendLinkX = 0;
	pendLinkY = 0;
	isPendingLink = false;
    }

    //both ends of link are now designated
    //
    void startLink (int x, int y)
    {
	// System.out.println ("Entered 'startLink()'");

	//store mouse loc for pending link.  If user goes through with
	//link creation, we use this for location info.
	pendLinkX = x;
	pendLinkY = y;
	isPendingLink = true;

        askForPredicate();
    }

    // finish creation of a new link, after dialog used to get name for it.
    //
    void createLink (String predStr, int linkArity)
    {
	/*  DEBUG
              System.out.println ("Inside 'createLink()'");
              System.out.println ("     predStr is:  " + predStr);
	*/

        // create new node, pass in this component & coordinates
        Link link = new Link (this, linkArity, predStr, pendLinkX, pendLinkY);
	isPendingLink = false;

        // Make the link the active object;  Redraw
        link.selectObject();
    }

    // either create new link, or rename existing link
    //
    void doLinkAction (String  selectedItem)
    {
/* TEST
	System.out.println ("Entered 'doLinkAction()' with selection: "
                            + selectedItem);
*/
	int  linkArity = parent.pn.getPredicateArity (selectedItem);

	if (isPendingLink)
            createLink (selectedItem, linkArity);
	else      // just a link modify
	    modifyLink (selectedItem, linkArity);

	// here the mouse up button which usually causes this does not
	// register, since a dialog pops up.  So we call it.
	//
	tb.setCurrentTool (Tools.MOVE);  // resets it unless previously
                                         // locked by a double click
        repaint();
    }

    // change data in an existing Link.  Is expected to be the
    // selected object now.
    //
    void modifyLink (String itsnewname, int itsarity)
    {
	NodeObject ob = (NodeObject)GraphObject.getActiveOb();
	if ((ob instanceof Link) && (ob != null))
	{
	    Link  lk = (Link) ob;
	    lk.changeData (itsnewname, itsarity);
            //lk.selectObject();  --call not needed.
	}
    }

    Node createNode (int x, int y)
    {
        // create new node, pass in this component & coordinates
        Node node = new Node (this, x, y);

        // Make the Node the active object
        node.selectObject();

	// Prepare for a rename action on current Node 
	tb.activateTextField();
	repaint();

        return node;
    }

    void cancelEdge()
    {
	initTempEdgeData();
	tb.setCurrentTool (Tools.MOVE);   //need because there is a dialog box,
	     //so mouse up event is not noticed by drawArea.
    }

    void initTempEdgeData()
    {
	isPendingEdge = false;
	pendingXPos = 0;
	mouseDragX = mouseDragY = 0;
    }

    void startEdge (int x, int y)
    {
	GraphObject  go = hitTest (x, y);
	if (go  instanceof  NodeObject)
	{
	    isPendingEdge = true;
	    go.selectObject();
	    pendingXPos = x;
	    mouseDragX = x;
	    mouseDragY = y;
	}
    }

    void finishEdge (int x, int y)
    {
	if (isPendingEdge)
	{
	    NodeObject ob = (NodeObject)GraphObject.getActiveOb();
	    GraphObject go = hitTest (x, y);

	    Node  node = null;
	    Link  link = null;
	    boolean  tonodeDir = true;
	    boolean  ignoreEdge = false;
	    int   argNum = 0;

	    if ((ob  instanceof  Node) && (go  instanceof  Link))
	    {
		node = (Node)ob;
		link = (Link)go;
		tonodeDir = false;
	        argNum = link.getArgNum (x);
		    //want the mouse up x in this case;
		    //need to use the particular mouse x value that is
		    //on the Link.
	    }
	    else  if ((ob  instanceof  Link) && (go  instanceof  Node))
	    {
		link = (Link)ob;
		node = (Node)go;
		tonodeDir = true;
	        argNum = link.getArgNum (pendingXPos);
		    //want the mouse down x in this case;
		    //need to use the particular mouse x value that is
		    //on the Link.
	    }
	    // if no Node exists where Edge is dropped, create an arb.
	    // named one, and connect Edge to it
            //
	    else  if ((ob  instanceof  Link) && (go == null))
	    {
		link = (Link)ob;
		node = createNode (x, y);
		tonodeDir = true;
	        argNum = link.getArgNum (pendingXPos);
		    //want the mouse down x in this case too
	    }
	    else
	        ignoreEdge = true;

	    Edge   pendingEdge;
	    if (!ignoreEdge)
                pendingEdge = new Edge (this, link, node, tonodeDir, argNum);
	    repaint();

	    initTempEdgeData();
	}
    }

    // Used to toggle Node  value when the check box changes
    public void toggleVarStatus (boolean newstate)
    {
	GraphObject  go = GraphObject.getActiveOb();

        if ((go != null) && (go instanceof Node))
        {
            changeVariable (newstate, false);
            repaint();
        }
    }

    public boolean mouseEnter (Event evt, int x, int y)
    {
        tb.resetCursor();
	    //make sure correct cursor is drawn for
	    //current tool
        return true;
    }

    public boolean mouseExit (Event evt, int x, int y)
    {
        parent.setCursor (Frame.DEFAULT_CURSOR);
        return true;
    }

    // handleEvent() handles the scrollbar events.
//TEST  -- delete or revise this comment later....
    // It updates the offset_x and offset_y variables that
    // are used by the canvas.

    //toggle the Variable item/button with the spacebar
    //delete graph objects with the delete key or backspace key
    //rename graph objects (active one) with the return/enter key
    //
    public boolean keyDown (Event evt, int key)
    {
        //System.out.println("ASCII is " + key);

        if (evt.id == Event.KEY_PRESS)
	{
	    /* In ascii decimal, here are the reps:
               ESCAPE     == 27
               BACK_SPACE == 8
               DELETE     == 127
               NEWLINE    == 10
               RETURN     == 13    (ENTER)
             */ 
            // Delete currently active object
	    if ((evt.key == Event.DELETE) || (evt.key == Event.BACK_SPACE))
		deleteObject();

            else  if (GraphObject.getActiveOb() instanceof Node)
            {
                // Toggle variable status
		if (evt.key == 32)  // a space was pressed
                {
                    changeVariable (false, true);
                    return true;
		}

		// Rename currently active Node
		else  if ((evt.key == Event.ENTER)  // key == 13
				|| (evt.key == 10) || (evt.key == 13))
		    tb.activateTextField();
            }
            else  if (GraphObject.getActiveOb() instanceof Link)
            {
		if ((evt.key == Event.ENTER)  // key == 13
				|| (evt.key == 10) || (evt.key == 13))
		    renameObject();
            }
	}
	return false;
    }

    // Java only recognizes one mouse button (the left one), for compatibility
    // with systems with only one button.  Others with multiple buttons
    // still generate mouse events, but Java sees the event as the mouse 1
    // key plus various modifiers.  The mouse buttons therefore generate
    // the following in an Event record:
    //
    // Mouse 1 button (left):    a mouse button without modifiers
    // Mouse 2 button (middle):  a mouse button plus Event.ALT_MASK
    // Mouse 3 button (right):   a mouse button plus Event.META_MASK
    //
    public boolean mouseDown (Event evt, int x, int y)
                       // mouse click position is (x,y)
    {
	//System.out.println ("Mouse Down coords  x, y:  " + x + ", " + y);

        int  which_mouse = 0;  //unset

	// "real" position, from offset of DrawArea (related to
	// scrollbar value).  (0, 0) is with respect to the current
	// "view" over the underlying component, given by the scrolling.
	//
        x = x + offset_x;
        y = y + offset_y;

	// calculate which mouse button was pushed (or equivalent)
	//
	which_mouse = MouseHelper.which_mouse_button(evt);
            //System.out.println ("which mouse is DOWN: " + which_mouse);

        switch (tb.getCurrentTool())
        {
            case   Tools.MOVE:

                // any mouse button down
		activateObject (x, y);

		/* ADDED hook for link descriptions... */
		if (evt.clickCount == 2)
		{
		    // find which object got hit (same as one just
                    //   activated, if any)
		    GraphObject obj = GraphObject.getActiveOb();
		    
		    // Set offset if it is an icon
		    if (obj instanceof Link)
		    {
			parent.handleLinkDoubleClick((Link) obj);
		    }
		}

                break;

            case   Tools.NODE:
		if (which_mouse == 1)
		{
                    // create new node
                    createNode (x, y);
		}
		else  if (which_mouse == 4)
                //button 1 with the shift or control key down
		//(if one of these is bound to a function the window manager
		//is using the application won't get it.  Hopefully, this
		//will allow the other to work.)
		{
		    // rename this node
                    setClass ("parka.Node");
                        //tell renameObject() that only Nodes are valid to ren.
		    renameObject (x, y);
		}
		else  if (which_mouse == 2)
		{
		    // delete this node
                    setClass ("parka.Node");
                        //tell deleteObject() that only Nodes are valid to del.
		    deleteObject (x, y);
		}
		else  if (which_mouse == 3)
		{
		    // does same thing as button 1 & shift or control key
                    setClass ("parka.Node");
                        //tell renameObject() that only Nodes are valid to ren.
		    renameObject (x, y);
		}
		else  ; //noop

                break;

            case   Tools.LINK:

		if (which_mouse == 1)
		{
		    startLink (x, y);
		}
		else  if (which_mouse == 4)  //button 1 with the shift key down
		{
		    // rename this node; this amounts to a delete/create new.
		    // note that this means any existing edges into this
		    // link are also deleted, and those are not put back.
		    // If new link is a diff arity, it's not clear where
		    // to put them.
		    //
                    setClass ("parka.Link");
                        //tell renameObject() that only Links are valid to ren.
		    renameObject (x, y);
		}
		else  if (which_mouse == 2)
		{
		    // delete this node
                    setClass ("parka.Link");
                        //tell deleteObject() that only Links are valid to del.
		    deleteObject (x, y);
		}
		else  if (which_mouse == 3)
		{
		    // does same thing as button 1 & shift or control key
                    setClass ("parka.Link");
                        //tell renameObject() that only Links are valid to ren.
		    renameObject (x, y);
		}
		else  ; //noop

                break;

            case   Tools.EDGE:

		if (which_mouse == 1)
		{
		    startEdge (x, y);
		}
		else  if (which_mouse == 4)  //button 1 with the shift key down
		{
		    // currently a no-op
		}
		else  if (which_mouse == 2)
		{
		    // delete this node
                    setClass ("parka.Edge");
                        //tell deleteObject() that only Links are valid to del.
		    deleteObject (x, y);
		}
		else  if (which_mouse == 3)
		{
		    // currently a no-op
		}
		else  ; //noop

                break;

            default:
                return false;
        }

        return true;
    }

    public boolean mouseUp (Event evt, int x, int y)
    {
	// "real" position, from offset of DrawArea (related to
	// scrollbar value).
	//
        x = x + offset_x;
        y = y + offset_y;

        switch (tb.getCurrentTool())
        {
            case Tools.NODE:
		break;

            case Tools.LINK:
                break;

            case   Tools.EDGE:

		// calculate which mouse button was pushed (or equivalent)
		// (only care for EDGE tool, unless we add more functions)
		//
		int which_mouse = MouseHelper.which_mouse_button(evt);
                        //System.out.println ("which mouse is UP " +
                        //     "(only note with EDGE tool): " + which_mouse);

		if (which_mouse == 1)
		{
		    finishEdge (x, y);
		}

                break;

            case Tools.MOVE:
                break;

            default:
    	        System.out.println ("ERROR!:  Expected another tool to exist");
        }        // end 'switch'

	// Must set this in mouse up, not mouse down, since some mouse
	// down actions require mouse drags to continue their actions,
	// and this already resets the current tool to the wrong icon in
	// that case.
	//
	tb.setCurrentTool (Tools.MOVE);  // resets it unless previously
                                         // locked by a double click
        return true;
    }
  
    public boolean mouseDrag (Event evt, int x, int y)
    {
        switch (tb.getCurrentTool())
        {
            // Move the Node around
            case  Tools.MOVE:
                    //any mouse button
		    GraphObject  go = GraphObject.getActiveOb();
		    if (go instanceof NodeObject)
		    {
        		x = x + offset_x;
        		y = y + offset_y;
		        ((NodeObject)go).moveGO (x, y);
                              // orig x and y are mouse coords.
		    }
                    break;

            case  Tools.NODE:
                    break;

            case  Tools.LINK:
                    break;

            // Draw the link line
            case   Tools.EDGE:
		    //should be 1st mouse button only
		    int which_mouse = MouseHelper.which_mouse_button(evt);
                           //System.out.println ("which mouse is DRAGGED " +
                           //     "(only note with Link tool): " + which_mouse);

		    if ((which_mouse == 1) && isPendingEdge)
		    {
			mouseDragX = x;
			mouseDragY = y;
			//redraw pending stuff
			repaint();
		    }
                    break;

            default:
    	         System.out.println ("ERROR!:  Another tool expected to exist");
        }
        return true;
    }

    void showPendingEdge (int x, int y)
    {
        GraphObject  gob = GraphObject.getActiveOb();
	if (gob instanceof NodeObject)
	{
	    NodeObject  no = (NodeObject)gob;

	    imageG.setXORMode (Edge.LINE_COLOR);

	    if (no  instanceof  Node)
	    {
		imageG.drawLine (x, y, ((NodeObject) no).X - getXSBoffset(),
                                       ((NodeObject) no).Y - getYSBoffset());
	    }
	    else  if (no  instanceof  Link)
	    {
		int   argPos = ((Link) no).getArgNum (pendingXPos);
		imageG.drawLine (x, y,
                                 ((Link) no).getXPos (argPos) - getXSBoffset(),
                                 ((Link) no).getYPos (argPos) - getYSBoffset());
            }
	    imageG.setPaintMode();
	    repaint();
	}
    }

    public GraphObject hitTest (int x, int y)
    {
        GraphObject obj = null;
      
	obj = Node.hitTest (x, y);
	if (obj == null)
	    obj = Link.hitTest (x, y);
	if (obj == null)
	    obj = Edge.hitTest (x, y);
/* TEST
	if (obj == null)
	    System.out.println ("NOTHING HIT\n");
	else
	    System.out.println ("Object hit is: " + obj.toString());
*/
	return   obj;
    }

    // returns true if it is a non-null NodeObject, false otherwise
    // however, any kind of GraphObject will be selected
    //
    boolean activateObject (int x, int y)
    {
	boolean  obtype = false;

	// find which object got hit
	GraphObject  temp = hitTest(x,y);

        // Make it the active object
	if (temp == null)
	{
	    unselectObject();
	}
	else
	{
            temp.selectObject();

            // Set offset if it is a node
            if (temp instanceof NodeObject)
	    {
                ((NodeObject)temp).setOffset(x,y);
                obtype = true;
            }
        }

        requestFocus();
	repaint();

	return  obtype;
    }

    // 'activeClass' guarantees that the active object has to be
    // the right type to be affected by the action
    //
    private boolean classValidate (Class c)
    {
        if (activeClass.equals(DEFAULT_CLASS))
            return  true;
        else   if (activeClass.equals(c))
            return  true;
        else
            return  false;
    }

    private void setClass (String newClassname)
    {
        try   {
            activeClass = Class.forName (newClassname);
        }
        catch  (ClassNotFoundException  nfe)
        {
            System.out.println ("Wanted class for " + newClassname);
            System.out.println (nfe.getMessage());
            activeClass = null;
        }
/* TEST
        if (activeClass == null)
            System.out.println ("New class is  NULL!");
        else
            System.out.println ("New class is:  " + activeClass.toString());
*/
    }

    // assumes an active object exists, and deletes it if it does,
    // unless its type (class) constraint is violated, if there is one.
    // (else does nothing)
    //
    boolean deleteObject()
    {
        boolean     success = true;

        GraphObject  go = GraphObject.getActiveOb();

        if ((go != null) && classValidate (go.getClass()))
        {
            go.deleteGO();
        }
        else
            success = false;

        unselectObject();
        repaint();

        // restore default, if not already set
        activeClass = DEFAULT_CLASS;

        return  success;
    }

    // deletes object made currently active one
    //
    boolean deleteObject (int x, int y)
    {
        boolean     success;

        GraphObject go = hitTest(x,y);
	if (go != null)
            go.selectObject();
            // Now 'go' is the active object

        success = deleteObject();

        return  success;
    }

    void renameObject()
    {
	GraphObject  go = GraphObject.getActiveOb();

        if (go != null)
        {
            // classValidate() guarantees that 'go' has to be the right type
            // to be affected by the action
            //
            if ((go instanceof  Node) && classValidate (go.getClass()))
            {
                // Prepare for an action on current Node
                tb.activateTextField();

                 String newStr = tb.getTextContents();
                 go.renameGO (newStr);
            }
            else   if ((go instanceof  Link) && classValidate (go.getClass()))
	    {
	         isPendingLink = false;
	         askForPredicate();
	    }
        }
        else
            unselectObject();

	repaint();

        // restore default, if not already set
        activeClass = DEFAULT_CLASS;
    }

    void renameObject (int x, int y)
    {
        GraphObject go = hitTest(x,y);
	if (go != null)
            go.selectObject();
            // Now 'go' is the active object

        renameObject();
    }

    // tells 'renameObject()' the type of object that is ok to rename
    // if type is "java.lang.Object", then any object is renamable
    //
    void renameObject (String obTypeToRename)
    {
        setClass (obTypeToRename);
		//param is full string name of class
        renameObject();    //this routine uses the currently active ob
    }

    // tells 'deleteObject()' the type of object that is ok to rename
    // if type is "java.lang.Object", then any object is renamable
    //
    void deleteObject (String obTypeToDelete)
    {
        setClass (obTypeToDelete);
        deleteObject();    //this routine uses the currently active ob
    }

    void changeVariable (boolean val, boolean togglep)
    {
	GraphObject  go = GraphObject.getActiveOb();

        if (go instanceof Node) 
        {
	    Node node = (Node)go;
	    node.changeVariable (val, togglep);

	    boolean newState = node.getIsVariable();
            tb.setVariableState (newState);
            parent.isvariableMenuItem.setState (newState);

            repaint();
        }
    }

    void updateFont (Font f, FontMetrics fm)
    {
        this.f = f;
        this.fm = fm;

	(getGraphics()).setFont (f);
	imageG.setFont (f);

	Node.updateFont (this);
	Link.updateFont (this);

        repaint();
    }

    void paintDrawArea()
    {
        repaint();
    }

    // determination that the list of bindings and list of variables is
    // valid is the responsibility of the caller.
    //
    void setGraphVariableBindings (String[] binds, String[] varnameList)
    {
        // clear all bindings
        Node.unbindVariables();

        // set bindings
        for (int i = 0; i < varnameList.length; i++)
        {
            Node  matching_var = Node.getVariable (varnameList[i]);
            if (matching_var != null)
            {
                matching_var.changeBinding (binds[i]);
            }
        }

        repaint();
    }

    void unsetGraphVariableBindings()
    {
        // clear all bindings
        Node.unbindVariables();
	repaint();
    }

    void noSelectionMode()
    {
	nodeDeselectionMode();
        parent.setMenuSensitivity (false, false, false);
        parent.pn.displayNumAnswersToQueryMsg();
    }

    void nodeDeselectionMode()
    {
        tb.deactivateNodeDisplays();
    }

    // set text field to string in new node; set "variable" button
    // display status
    //
    void nodeSelectionMode (Node n)
    {
	tb.activateNodeDisplays (n);
    }

    void linkSelectionMode()
    {
    }

    void toggleMenuStatus (boolean sensitizeObjMenu, boolean activeItemState)
    {
        //obj. menu becomes sensitive; link menu is made insensitive
	if (sensitizeObjMenu)
	    parent.setMenuSensitivity (true, false, activeItemState);
	else
	    parent.setMenuSensitivity (false, true, false);
	    //activeItemState is ignored; only relevant for a Node
    }

    public void printDebug()
    {
	Node.printDebug();
	Link.printDebug();
	Edge.printDebug();
    }
}
