/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
/**********************************************************************
  edges -- connect frames and predicates.
  **********************************************************************/
package parka;

import java.awt.*;
import java.util.*;

public class Edge extends parka.GraphObject
{
    // all edges in this graph
    //
    static  Vector edgesVector = new Vector();

    // the two ends of the line
    private Link  linkend;
    private Node  nodeend;

    private boolean   toNode;    //direction of line
	// toward node == true;
	// toward link == false
    private int     argPosition; //won't change; user has to delete
	//edge and redraw it to get a new value
	// Value range:  1 to arity of linkend.

    private Color c;

    // color
    final static Color LINE_COLOR = Color.blue;
    final static Color RANGE_COLOR = Color.blue;
    final static Color HI_COLOR = Color.red;

    final static int ARROW_RADIUS = 3;
    final static int EPSILON_SQUARED = 4;

    // *******
    // Methods
    // *******

    Edge (DrawArea MyDraw, Link l, Node n, boolean toNodeDir, int argnumber)
    {
        super (MyDraw);

	if (l.isEmptyArg(argnumber))
	{
	    linkend = l;
	    nodeend = n;
	    toNode = toNodeDir;
	    argPosition = argnumber;
	    if (toNode)
                c = RANGE_COLOR;
	    else
                c = LINE_COLOR;

	    l.addNode (n, this, argnumber);
	    n.addLink (l, this); //addLink call installs info into adjacent node

	    // Uses 'draw' variable in ancestor class to remember its DrawArea.

	    // Add the new Edge to edgesVector
	    edgesVector.addElement (this);
	}
	// else, let new Edge garbage collect, because no one refs it.
    }

    Node getNode()
    {
	return  nodeend;
    }

    Link getLink()
    {
	return  linkend;
    }

    // Do in the reverse order that they were drawn so you select
    // top objects first
    //
    static Edge hitTest (int x, int y)
    {
        for (int i=(edgesVector.size()-1); i >= 0; i--)
        {
            Edge obj = (Edge) edgesVector.elementAt(i);
	    /* System.out.println ("Edge hitTest:  obj ends are -- "
                         + obj.linkend.str + "  " + obj.nodeend.str);   */

            if (obj.mouseWithin (x,y))
                return obj;
        }
	return   null;
    }

    // If the distance of mouseX and mouseY is within epsilon of the
    // closest realX and realY on the line representing this Edge, then
    // we have a HIT.
    //
    boolean mouseWithin (int mouseX, int mouseY)
    {
/* TEST
	System.out.println ("Edge.mouseWithin():   (mouse x, mouse y) == " +
             mouseX + ", " + mouseY);
*/
	// keep knowledge of how Link is drawn away from Edge:
	int    linkXConnect = linkend.getXPos (argPosition);
        int    linkYConnect = linkend.getYPos (argPosition);
	       //
	       //we don't keep this knowledge away from Edge for Node; it
	       //currently makes no difference.  But for strict generality,
	       //we should.  What if its shape changes?

        if (nodeend.X != linkXConnect)
        {
	    float  slope;

            if (toNode)
                slope = ((float)(nodeend.Y - linkYConnect)) /
                        ((float)(nodeend.X - linkXConnect));
            else
                slope = ((float)(linkYConnect - nodeend.Y)) /
                        ((float)(linkXConnect - nodeend.X));

/* TEST
	    System.out.println ("   Node (x,y): (" + nodeend.X + ", " +
                 nodeend.Y + ")    Link (x, y): (" + linkXConnect +
                 ", " + linkXConnect + ")");
*/

	    if (linkYConnect != nodeend.Y)  //slope != 0
	    {
		// y = slope*x + b
		float   b = nodeend.Y - (slope * nodeend.X);

                // slope of perpendicular line:
		float   negInverseSlope = -( 1/slope);
		// b for the perpendicular line:  (we pick the perpendicular
                //    line containing the mouse click point)
		float   inverseB = mouseY - (negInverseSlope * mouseX);

		// point where both lines intersect has point
		// (interceptX, interceptY).  Thus, if subtract equation for
		// perpendicular line from the orig. line, we get:
		// 0 = (slope - negInverseSlope) * interceptX + (b - inverseB).
		// Therefore:
		float interceptX = (inverseB - b) / (slope - negInverseSlope);
		float interceptY = slope * interceptX + b;
		//
		// System.out.println ("   (interceptX, interceptY) ==  "
			// + interceptX + ", " + interceptY);

		float distance = (mouseX - interceptX) * (mouseX - interceptX)
                           + (mouseY - interceptY) * (mouseY - interceptY);
		// System.out.println ("   distance (squared) = " + distance);
		if (distance <= EPSILON_SQUARED)
		    return  true;
	    }
	    else             //slope == 0
	    {
                if (linkXConnect < nodeend.X)
                {
		    if ( (linkXConnect < mouseX) && (mouseX < nodeend.X) &&
                         ((mouseY - nodeend.Y) * (mouseY - nodeend.Y) <=
                                    EPSILON_SQUARED) )
		         return  true;
                }
                else  //nodeend.X <= linkXConnect
                {
		    if ( (linkXConnect > mouseX) && (mouseX > nodeend.X) &&
                         ((mouseY - nodeend.Y) * (mouseY - nodeend.Y) <=
                                    EPSILON_SQUARED) )
		         return  true;
                }
	    }
        }
        else   //nodeend.X == linkXConnect  -- infinite slope
        {
            if (linkYConnect < nodeend.Y)
            {
		if ( (linkYConnect < mouseY) && (mouseY < nodeend.Y) &&
                     ((mouseX - nodeend.X) * (mouseX - nodeend.X) <=
                                EPSILON_SQUARED) )
		     return  true;
            }
            else  //nodeend.Y <= linkYConnect
            {
		if ( (linkYConnect > mouseY) && (mouseY > nodeend.Y) &&
                     ((mouseX - nodeend.X) * (mouseX - nodeend.X) <=
                                EPSILON_SQUARED) )
		     return  true;
            }
        }
        return  false;
    }

    static void drawEdges (Graphics g)
    {
        Enumeration e;
	Edge        edge;

        e = Edge.edgesVector.elements();
        while (e.hasMoreElements())
        {
            try
            {
                edge = (Edge)e.nextElement();
                edge.drawGO (g);
            }
            catch (NoSuchElementException e1)
            {
                // Should not get here.
		System.out.println ("Exception: No such edge");
            }
        }
    }

    void drawGO (Graphics g)
    {
	int  xoff = draw.getXSBoffset();
	int  yoff = draw.getYSBoffset();

	int  beginx, beginy, endx, endy;

	if (toNode)
	{
	    beginx = linkend.getXPos (argPosition) - xoff;
	    beginy = linkend.getYPos (argPosition) - yoff;
	    endx = nodeend.X - xoff;
	    endy = nodeend.Y - yoff;
	}
	else
	{
	    beginx = nodeend.X - xoff;
	    beginy = nodeend.Y - yoff;
	    endx = linkend.getXPos (argPosition) - xoff;
	    endy = linkend.getYPos (argPosition) - yoff;
	}

	g.setColor (c);
	g.drawLine (beginx, beginy, endx, endy);

        if (toNode)   //then draw a nob on the 'to' end.
	{
	    drawNob (g, (float)beginx, (float)beginy, (float)endx, (float)endy);
	}
	else
	{
	    drawNob (g, (float)endx, (float)endy, (float)beginx, (float)beginy);
	}
    }

    // Calculate location of arrow by finding the intersection
    // of the line and box
    // Calling this function implies we are drawing the oval on
    // a node.
    //
    private void drawNob (Graphics g, float x1, float y1, float x2, float y2)
    {
        float xx = 0, yy = 0;

        // Check intersection with the top or bottom
        if (y1 != y2) {
            yy = y2 - (nodeend.height / 2) - nodeend.YBUF;
            if (y2 < y1)  
                yy += nodeend.height + (2 * nodeend.YBUF);
            if (x1 == x2)  
                xx = x2;
            else
                xx = ((x2-x1)/(y2-y1)) * (yy-(((x2*y1) - (x1*y2))/(x2-x1)));
        }
    
        // If not, check for left or right side
        if ((xx < (x2 - (nodeend.width / 2) - nodeend.XBUF)) ||
             (xx > (x2 + (nodeend.width / 2) + nodeend.XBUF))) 
        {
            xx = x2 - (nodeend.width / 2) - nodeend.XBUF;
            if (x2 < x1) 
                xx += nodeend.width + (2 * nodeend.XBUF);
            if (y2 == y1)
                yy = y2;
            else
                yy = (((y2 - y1) / (x2 - x1)) * xx) +
                                       (((x2 * y1) - (x1 * y2)) / (x2 - x1));
        }
    
        // Draw the arrow
        g.fillOval ((int)xx-ARROW_RADIUS , (int)yy-ARROW_RADIUS,
                    ARROW_RADIUS * 2, ARROW_RADIUS * 2);
    }

    void deleteLink()
    {
	linkend = null;
    }

    void deleteNode()
    {
	nodeend = null;
    }

    // Delete the edge;  used when edge is object chosen to delete
    void deleteGO()
    {
	linkend.deleteEdge(this);
	//nodeend will delete from the above call too.
	//both 'deleteLink()' and 'deleteNode()' will be called from
        //'deleteEdge()'

	edgesVector.removeElement (this);
    }

    // For selecting and highlighting the GraphObject
    public void selectGO()
    {
	c = HI_COLOR;
	active = this;
    }

    // For deselecting the GraphObject
    public void unselectGO()
    {
        if (toNode)
            c = RANGE_COLOR;
        else
            c = LINE_COLOR;

	active = null;
    }

    static void clearAll()
    {
	edgesVector.removeAllElements();
    }

    public void debug(String str)
    {
        System.out.println("");
        System.out.println (str + "Edge: ");
        System.out.print ("  " + str + " " + linkend.str + " ; ");
        System.out.println (str + " " + nodeend.str);
    }

    public static void printDebug()
    {
        System.out.println("");
        System.out.println (Edge.edgesVector.size() + " links");
        for (int i=0; i < Edge.edgesVector.size(); i++)
        {
	    Edge e = (Edge)edgesVector.elementAt(i);
	    e.debug(" ");
	}
    }
}
