/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
/**********************************************************************
  GraphObject
  **********************************************************************/
package parka;

import java.awt.*;
import java.util.*;

abstract class GraphObject extends java.lang.Object
{
    // draw area pointer
    protected    DrawArea draw;
    protected    static GraphObject   active;   //currently selected object

    // *******
    // Methods
    // *******

    GraphObject (DrawArea  da)
    {
	super();
	draw = da;
    }

    // Draw the GraphObject
    abstract void drawGO (Graphics g);

    // Delete the GraphObject
    abstract void deleteGO();

    // Determine if the coordinate is in the GraphObject
    abstract boolean mouseWithin (int x, int y);

    // For selecting the GraphObject
    abstract void selectGO();

    // For deselecting the GraphObject
    abstract void unselectGO();

    // For multiplexing the select/unselect

    // Does all the work of updating the window when a new GraphObject
    //    (or null) is selected.
    //
    void selectObject()
    {
        // unhighlight currently selected graphObject
        if (active != null)
	{
	    /*System.out.println ("GraphObject.selectObject():  " +
                               "active is NOT NULL");   */
            active.unselectGO();
	    draw.noSelectionMode();  // reinitialize for any object type
	}
	/* else
	    System.out.println ("GraphObject.selectObject():  " +
                               "active is NULL");  */

	selectGO();

        // Redraw 
        draw.paintDrawArea();
    }

    // use when no object is to be selected
    //
    static void unselectObject (DrawArea da)
    {
        // unhighlight currently selected graphObject
        if (active != null)
	{
            active.unselectGO();
	    active = null;
	}

        // Nothing selected
	da.noSelectionMode();

        // Redraw 
	da.paintDrawArea();
    }

    public static  GraphObject getActiveOb()
    {
	return  active;
    }

    void renameGO (String newstr) {}
    //subclasses using this must override

    // Debugging function
    abstract void debug (String str);
}
