/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
package parka;

import java.awt.*;

public class KBDialog extends ParkaDialog
{
    NetKBInit   kbInit;

    // Button constants
    public static final int CANCEL = 0;
    public static final int OK = 1;

    protected static final String TITLE = "Please select a Knowledge Base";
    static final String  MEM = "Memory";
    static final String  DISK = "Disk";
    
    protected List list;
    protected CheckboxGroup cbGroup;

    /*******
     Methods
     *******/
    
    public KBDialog (ParkaWin parent, NetKBInit initOb, String[] kbs)
    {
        super (parent, TITLE);

	kbInit = initOb;

        // Specify a LayoutManager for it
        setLayout(new BorderLayout(5, 5));

        // Set up list
        list = new List(5, false);
	for (int i = 0; i < kbs.length; i++)
            list.addItem (kbs[i]);
        list.select(0);  // preselect first item

        // Put the list
        this.add("Center", list);

        // Create a panel of buttons, center the row of buttons in
        // the panel, and put the pane at the bottom of the window.
        Panel p = new Panel();
        p.setLayout (new FlowLayout(FlowLayout.CENTER, 15, 15));
        p.add (ok = new Button (OKlabel));
        p.add (cancel = new Button (CANCELlabel));
        this.add ("South", p);

        // Create Memory/Disk check box group
        cbGroup = new CheckboxGroup();
        Panel p2 = new Panel();
        p2.setLayout (new FlowLayout(FlowLayout.CENTER, 0, 0));
        p2.add (new Checkbox (MEM, cbGroup, true));
        p2.add (new Checkbox (DISK, cbGroup, false));
        add ("North", p2);

        // Set the window to its preferred size.
        this.pack();
    }

    // Handle button events by calling the answer() method.
    // Pass the appropriate constant value, depending on the button.
    public boolean action (Event e, Object arg)
    {
	return  (super.list_action (e, arg, list));
    }

    private boolean useMemOrDisk()
    {
	Checkbox whichActive = cbGroup.getCurrent();
	String   activeLabel = whichActive.getLabel();
	if (activeLabel.equals (MEM))
	    return  true;
	else  if (activeLabel.equals (DISK))
	    return  false;
	else
	{
	    System.out.println ("Unrecognized option for KB characteristic. "
                    + "Internal memory used.");
	    return  true;
	}
    }

    // Do Ok action
    protected void ok()
    {
	if (kbInit == null)
	{
	    System.out.println ("Programmer Error:  Class for loading " +
                  "KB does not exist");
	    return;
	}
        /* System.out.println ("In OK-- KBname is: [" + list.getSelectedItem()
                       + "] ; useMem = " + useMemOrDisk());
        */
        kbInit.loadKBbyName (list.getSelectedItem(), useMemOrDisk());
    }
}
