/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
/**********************************************************************
  link
  **********************************************************************/
package parka;

import java.awt.*;
import java.util.*;

public class Link extends parka.NodeObject
{
    private int     arity;

    // all links in this graph
    //
    static  Vector  linksVector = new Vector();

    private Node    arguments[];
		// the 0th position is the 'from' Node; the rest, in order
		// are the arguments: 1st position is 2nd arg (first 'to'
		// node, etc.

    private Vector  edgelist;

    private static String  emptyArg = " ";

    Color c;

    // color
    final static Color HI_COLOR = Color.yellow;

    final static   int   RECTWD     = 8;   //in pixels
    final static   int   RECTHT     = 14;  //in pixels

    // *******
    // Methods
    // *******

    Link (DrawArea MyDraw, int argnum, String s, int x, int y)
    {
        super (MyDraw, x, y);

	arity = argnum;
	arguments = new Node[arity];
	edgelist = new Vector();

        str = s;
        c = DEF_COLOR;

	// uses 'draw' variable in ancestor class to remember its DrawArea
	// Add new link to the static list linksVector
	//
	linksVector.addElement (this);

    //  's' is the link name, and is chosen using the PredDialog class
    }

    void changeData (String newS, int newarity)
    {
	str = newS;

	Node temparray[] = arguments;
	arguments = new Node[newarity];

	// if arity is bigger, the extra edges are dropped;
	// if newarity is bigger, the extra positions are null
	//
	if (arity > newarity)
	    System.arraycopy (temparray, 0, arguments, 0, newarity);
	else
	    System.arraycopy (temparray, 0, arguments, 0, arity);

        if (arity > newarity)   //arguments.length == arity
	{
//XXX
System.out.println ("old arity greater than new arity; old: " + arity
                    + " ; newarity: " + newarity);

	    //throw out the extra edges:
	    for (int i = newarity; i < arity; i++)
	    {
	        Edge  edge = matchEdge ( ((Node)temparray[i]).getObjName() );
	        if (edge != null)
		{
// XXX  TEST
System.out.println ("removing:  " + edge.toString());
		    edge.deleteGO();
		}
	    }
	}

	arity = newarity;
    }

    private Edge matchEdge (String  nodeNameToMatch)
    {
//XXX
System.out.println ("Link.matchEdge() called");

	for (int i = 0; i < edgelist.size(); i++)
	{
	    Edge  e = (Edge)(edgelist.elementAt(i));
	    String  edgename = e.getNode().getObjName();
	    if (edgename.equals (nodeNameToMatch))
		return   e;
	}
	return   null;
    }

    // Do in the reverse order that they were drawn so you select
    // top objects first
    //
    static Link hitTest (int x, int y)
    {
        for (int i=(linksVector.size()-1); i >= 0; i--)
        {
            Link  obj = (Link) linksVector.elementAt(i);

            if (obj.mouseWithin (x,y))
                return obj;
        }
	return   null;
    }

    boolean mouseWithin (int x, int y)
    {
/* TEST
	System.out.println ("Link.mouseWithin()   x,y:  " + x + ", " + y);
	System.out.println ("            X,Y:  " + X + ", " + Y);
*/
	int  ychange = RECTHT/2;
	int  xchange = (arity * RECTWD)/2;

        return ((x > X - xchange) && (x < X + xchange) &&
                (y > Y - ychange) && (y < Y + ychange));
    }

    // it is assumed that it has been determined that 'x' mouse position
    // is within the graphic for this link.  Here it is determined, by
    // position, which argument it is.  If the result is not within the
    // arity for this link, 0 is returned.
    //
    int getArgNum (int x)  //mouse x position
    {
	int  argnum = 1;
	int  xleft = X - (arity * RECTWD)/2;   //no offset in calc:
		//figure has already taken scrollbar offset into account
	int  localX = x - xleft;

/* TEST
	System.out.print ("getArgNum(): X == " + X + "; xleft == " + xleft);
	System.out.println ("  localX == " + localX);
*/
	if (localX < 0)
	    return  0;

	while (localX > argnum * RECTWD)
	    argnum++;

	//System.out.println ("    getArgNum(): arg position is  " + argnum);

	if (argnum > arity)
	    return  0;
	else
	    return  argnum;
    }

    // x for object, given which argument we have.  We need to position
    // the x inside the right sub-rectangle.
    // Figures position for an already drawn object (so that scrollbar
    // offset is already handled).
    //
    int getXPos (int argPosition)
    {
	int  xleft = X - (arity * RECTWD)/2;
	int  argXOffset = (argPosition -1) * Link.RECTWD + (Link.RECTWD/2);
                         //controls which rect inside Link the linkend is in
/* TEST
	System.out.println ("  getXPos():  Arg pos == " + argPosition +
                    "; argXOffset == " + argXOffset + "; xleft == " + xleft);
*/
	return  (xleft + argXOffset);
    }

    // Center of object (up-down-wise)
    // Figures position for an already drawn object (so that scrollbar
    // offset is already handled).
    // We put the argument 'argPosition' here, so that if the representation
    // changes (eg a 90 degree change in how the link is drawn), the caller
    // doesn't need to know which position, x or y or both needs the arg info.
    //
    int getYPos (int argPosition)
    {
	return  (Y);
    }

    static void drawLinks (Graphics g)
    {
        Enumeration e;
	Link        link;

        e = Link.linksVector.elements();
        while (e.hasMoreElements())
        {
            try
            {
                link = (Link)e.nextElement();
                link.drawGO (g);
            }
            catch (NoSuchElementException e1)
            {
                // Should not get here.
		System.out.println ("Exception: No such link");
            }
        }
    }

    void drawGO (Graphics g)
    {
	int  xoff = draw.getXSBoffset();
	int  yoff = draw.getYSBoffset();
	int  xleft, ytop, xline;

	// draw rectangles
	int  totalWid = arity * RECTWD;

/* TEST
	System.out.print ("Link.drawGO():  (X, Y) == (" + X + ", " + Y);
	System.out.println ("); " + "   totalWid == " + totalWid);
*/
	ytop = Y - RECTHT/2 - yoff;
	xleft = X - totalWid/2 - xoff;

	// System.out.println (";   xleft == " + xleft + "; ytop == " + ytop);

	Color  heldc = g.getColor();
        g.setColor (c);                      // set background color
	g.fillRect (xleft, ytop, totalWid, RECTHT);
        g.setColor (Color.black);
	g.drawRect (xleft, ytop, totalWid, RECTHT);

	// xline = xleft;
	for (int i = 0; i < arity -1; i++)
	{
	    // now draw vertical lines to make rectangles within the rect
	    xleft += RECTWD;
	    g.drawLine (xleft, ytop, xleft, ytop + RECTHT);
	}

        // draw string, above
	// int   xtext = xline;   --use if left side of text lines up
	//                       with left side of Link icon
	width = resetWidth();
	height = resetHeight();

	int   xtext = X - width/2;      // center text over Link icon
	int   ytext = ytop - descent;

        g.setColor (Color.black);   // use black ink
	g.drawString (str, xtext, ytext);
	g.setColor (heldc);         // reset color
    }

    static void updateFont (DrawArea da)
    {
	NodeObject.updateFont(da);

	Enumeration elts = linksVector.elements();
        while (elts.hasMoreElements())
        {
            Link l = (Link)elts.nextElement();
	    l.width = l.resetWidth();
	    l.height = l.resetHeight();
	}
    }

    // returns true if this argument position is empty (no Edge to
    // some Node from it
    //
    boolean isEmptyArg (int thisArg)
    {
	return   (arguments[thisArg -1] == null);
    }

    void addNode (Node n, Edge e, int argPosition)
    {
	if ((argPosition <= arity) && (argPosition > 0))
	{
	    arguments[argPosition -1] = n;
	    edgelist.addElement (e);
	}
	else
	{
	    System.out.println ("Programming Error:  the arity of "
               + "'addEdgeToLink()' requires that the position of ");
	    System.out.println ("  and argument must be between 1 and "
               + arity + ", inclusive.");
	}
    }

    // Return the link with the give name
    //
    public static Link getLink (String str)
    {
	// System.out.println ("       " + str);

        Enumeration e = linksVector.elements();
	Link    link;
        while (e.hasMoreElements())
        {
            link = (Link)e.nextElement();
	    // System.out.println ("       this link's name: <"
            //              + link.getObjName() + ">");

            if ((link.getObjName()).equals (str))
                return  link;
        }

        return null;
     }

    // Delete this link
    void deleteGO()
    {
	removeAllEdges();
        linksVector.removeElement (this);

/*      TEST
	System.out.println ("Leaving  deleteGO()");
	System.out.println ("  linksVector.toString:  " +
                              linksVector.toString());
*/
    }

    //should only be called from Node or Link's 'deleteEdge()' routine.
    protected void delete (Edge e, Node otherEnd)
    {
	for (int i = 0; i < arguments.length; i++)
	{
            //testing for same obj, not just same values with '=='
	    if (arguments[i] == otherEnd)
	    {
		arguments[i] = null;  //do NOT delete
		break;
	    }
	}
	for (int i = 0; i < edgelist.size(); i++)
	{
	    if ((Edge)edgelist.elementAt(i) == e)
	    {
		edgelist.removeElementAt(i);
                e.deleteLink();
		break;
	    }
	}
    }

    protected void deleteEdge (Edge e)
    {
	Node  otherEnd = e.getNode();

	/*System.out.println ("Connected node (to del. edge) is:  "
                                 + otherEnd.str);   */

	if (otherEnd != null)
	{
	    delete (e, otherEnd);
	    otherEnd.delete (e, this);
            Edge.edgesVector.removeElement (e);
	}
    }

    //removes Edge from edgelist and deletes it;
    //removes Node from arguments array
    //
    void removeAllEdges()
    {
        // here we use a clone of the Vector (which will be no problem,
        // because it's a shallow copy) in order to delete from the
        // edgelist without screwing up the internal pointers to items
        // when Enumerating
	//
	Vector  edgelistCopy = (Vector)edgelist.clone();
        Enumeration   elts = edgelistCopy.elements();
        while (elts.hasMoreElements())
        {
            Edge  e = (Edge)elts.nextElement();
	    deleteEdge (e);
        }
    }

    String getObjLabel()
    {
	return   str;
    }

    // For selecting and highlighting the GraphObject
    public void selectGO()
    {
        c = HI_COLOR;
	active = this;

        // deactivate text string and checkbox
	draw.nodeDeselectionMode();
	draw.toggleMenuStatus (false, false);
    }

    // For deselecting the GraphObject
    public void unselectGO()
    {
        c = DEF_COLOR;
	active = null;
    }

    Vector getLinkArguments()
    {
	Vector args = new Vector();

	for (int i = 0; i < arguments.length; i++)
	{
            StringBuffer str = new StringBuffer();

	    if (arguments[i] == null)
		str = str.append (emptyArg);
	    else
	    {
		if ((arguments[i]).getIsVariable())
		    str = str.append (String.valueOf (Node.VARIABLE_PREFIX));
		str.append ( ((Node)arguments[i]).getObjName());
	    }
	    args.addElement (new String(str));
	}

	return  args;
    }

    boolean isArgVariable(int pos)
    {
	return (arguments[pos]).getIsVariable();
    }

    Vector getLinkArgBindings()
    {
	Vector args = new Vector();

	for (int i = 0; i < arguments.length; i++)
	{
            StringBuffer str = new StringBuffer();

	    if (arguments[i] == null)
		str = str.append (emptyArg);
	    else
	    {
		str.append ( ((Node)arguments[i]).getBinding());
	    }
	    args.addElement (new String(str));
	}

	return  args;
    }

    void raiseSelected()
    {
	linksVector.removeElement (this);
	linksVector.addElement (this);
    }

    static void clearAll()
    {
        linksVector.removeAllElements();
    }

    public void debug (String str)
    {
        System.out.println ("");
        System.out.println (str + "Link: " + this.str);
	System.out.println (str + "Arguments (connected nodes, in order)");
	for (int i = 0; i < arguments.length; i++)
	{
	    Node  n = (Node)arguments[i];
	    if (n != null)
	        n.debug("  ");
	}
	System.out.println (str + "Edges for " + this.str);
	for (int i = 0; i < edgelist.size(); i++)
	    ( (Edge)edgelist.elementAt(i) ).debug("  ");
        System.out.println ("");
    }

    public static void printDebug()
    {
        System.out.println("");
        System.out.println (linksVector.size() + " links");
        for (int i=0; i < linksVector.size(); i++)
        {
            Link l = (Link)linksVector.elementAt(i);
            l.debug(" ");
        }
    }
}
