/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
package  parka;

import  java.io.*;
import  java.awt.*;

class  NetKBInit  extends Object
                  implements  Runnable
{
    Thread    KBloader;
    ParkaNet  pn;

        // KB information
    private String    KBName;
    private boolean   useMem;
    private InputStream   instream;   //expected to be set up elsewhere
    private OutputStream  outstream;  //expected to be set up elsewhere

    private int  oldCursor;

    NetKBInit (ParkaNet  pnet, InputStream is, OutputStream os)
    {
	pn = pnet;
	KBloader = null;
	KBName = "";
 	instream = is;
	outstream = os;
	oldCursor = 0;    // unset
    }

    boolean  isThreadAlive()
    {
	//we don't really care if it's started or not; just if it exists
	//
	if (KBloader != null)
	    return  true;
	else
	    return  false;
    }

    void join()
    {
	try  {
	    if (KBloader != null)
	        KBloader.join();
	}
	catch (InterruptedException ie)  {}
    }

    // Start Load a KB
    // With a GUI interface, this is the routine to call
    // in this class, after the constructor.  The work done
    // by this class starts from here.
    //
    void loadKB()
    {
	if (pn.isConnection())
        {
	    if (KBloader != null)
            {
	        try   {
		    KBloader.join();
                         //finish a diff KB load first, if one is going
                }
                catch  (InterruptedException  e)  {
                    System.out.println ("KB loader thread interrupted");
                }
            }
            KBloader = new Thread (this, "KBloaderThread");

            // Get list of KBs
            String[] kbs = getKBs();

            // Have user choose KB and memory/disk option
            KBDialog cbd = new KBDialog (pn.getWin(), this, kbs);
	    if (ApplType.getIsVersion1_0())
		((Window)cbd).show();
	    else
                cbd.show();
	}
	else
            (pn.getWin()).displayMessage (ParkaNet.NOCONNMSG);
    }

    // KB loading occurs in a separate thread, until done.
    public void run()
    {
	// System.out.println ("Entered 'run() (for thread to load KB)'");

        // load KB
        pn.openKB (KBName, useMem);   //this part may take a good while

	// this part should be quick with any KB:
        if (pn.isKBOpen())
        {
            // Restore old cursor
            (pn.getWin()).setCursor (oldCursor);

            (pn.getWin()).displayMessage("Finished Loading " + KBName);
	}
	// else message displayed from ParkaNet.openKB() is just fine.

	KBloader = null;
    }

    /* if dialog of kb names has popped up, and 'ok' is chosen, this routine
       will run, close an existing kb, if any, and open the chosen one.
       Waiting until here to close the existing kb allows the user to cancel
       without having the old kb data erased, if they change their minds.
    */
    void loadKBbyName (String kb, boolean useMem)
    {
	// System.out.println ("Entered 'loadKBbyName()'");

        // Set up wait cursor
        oldCursor = (pn.getWin()).getCursorType();
        (pn.getWin()).setCursor (Frame.WAIT_CURSOR);

        if (pn.isKBOpen())
            pn.closeKB();

	KBName = kb;
	this.useMem = useMem;
	/* System.out.println ("In 'loadKBbyName()': KBname is: [" + KBName +
                       "] ; useMem = " + useMem);
        */
	// need the test for null if 'loadKB()' is not called first;
	// (some uses of this class do not include using the dialog box
	// to get the kb name)
	//
	if (KBloader == null)
	    KBloader = new Thread (this, "KBloaderThread");
	KBloader.start();
    }

    private String[] getKBs()
    {
        (pn.getWin()).displayMessage("Retrieving KB list");
	String[]  kblist;

        // Returns a list of available Knowledge Bases
	if (pn.isConnection())
        {
            Command  c = new Command (pn, Command.GETKBNAMES, "");
	    c.sendCommand();
            Response r = new Response (pn);
	    r.computeResponse();
	    if (r.getStatus() > 0)
	    {
	        r.makeTokens();
	        kblist = r.getTokenList();
	    }
	    else
	    {
	        kblist = new String[0];
		(pn.getWin()).displayMessage (r.getResponse());
	    }
        }
	else
        {
	    (pn.getWin()).displayMessage (ParkaNet.NOCONNMSG);
	    kblist = new String[0];
        }

        return   kblist;
    }
}
