/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */

/**********************************************************************
  node object
  **********************************************************************/
package parka;

import  java.awt.*;

abstract class NodeObject extends parka.GraphObject
{
    // Coordinates
    protected int X;
    protected int Y;

    //derived from font info
    protected  int width;
    protected  int height;
    static  protected  int descent;

    private int dx;
    private int dy;

    // node or link name
    protected  String str;

    protected  final static Color DEF_COLOR = new Color(210,210,210);

    NodeObject (DrawArea MyDraw, int x, int y)
    {
	super (MyDraw);
        dx = 0;
        dy = 0;
	X = x;
	Y = y;
	str = "";

	updateFont (MyDraw);
        width = resetWidth();
	height = resetHeight();
    }

    // Save the offset distance between the mouse and the object coordinates
    void setOffset (int x, int y)
    {
        dx = x - X;
        dy = y - Y;
    }

    void moveGO (int x, int y)
    {
        X = x - dx;   // dx change in x offset
        if (X < 0)
            X = 0;
        if (X >= DrawArea.MAX_SIZE)
            X = DrawArea.MAX_SIZE-1;
        Y = y - dy;
        if (Y < 0)
            Y = 0;
        if (Y >= DrawArea.MAX_SIZE)
            Y = DrawArea.MAX_SIZE-1;

        draw.repaint();
	   //'draw' is in parent class
    }

    // Get text for the graphics
    abstract  String getObjLabel();

    String getObjName()
    {
        return   str;
    }

    int getX()
    {
	return  X;
    }

    int getY()
    {
	return  Y;
    }

    protected int resetWidth()
    {
	/* System.out.println (getObjLabel() + "  str width: "
               + (draw.getDAFontMetrics()).stringWidth (getObjLabel()) );
	 */
	return  ((draw.getDAFontMetrics()).stringWidth (getObjLabel()));
    }

    protected int resetHeight()
    {
	return  ((draw.getDAFontMetrics()).getAscent() + descent);
    }

    // Update the font and dependent information for the NodeObject
    //
    static  void updateFont (DrawArea  da)
    {
	FontMetrics fm = da.getDAFontMetrics();
        descent = fm.getDescent();
    }
}
