/* Parka Java Interface v. 1.0.2 (windowing version)
   Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

   Authors: Sean Luke, Shiloh Heurich, David Rager, Carolyn Gasarch,
            Jeff Heflin, Brian Sullivan
   Contact: Jeff Heflin (heflin@cs.umd.edu)
   */

/**********************************************************************
  standalone
**********************************************************************/
package parka;

import java.awt.*;
import java.util.*;
import java.io.*;
import java.net.URL;
import java.net.MalformedURLException;

public class ParkaApp extends java.awt.Frame
{
    final static int width = 600;
    final static int height = 500;
    final static String  QUIT = "Quit";
    final static String  NEWWIN = "New Window";
    final static String  ICON_FILE = "IconDirLoc";

    Vector WinV;
    String url_directory;
    static  ApplType apptype;

    // Allows autoloading of KB -
    String parkaHost = null;
    int parkaPort = 7990;
    String parkaKB = null;

    /* Note : there is a bug in jdk 1.1.4 and 1.1.5 where the titles
       do not show for either Frame or Dialog subclasses.  One comment
       under the Sun Developer connection says that running under 1.1.3
       if you compile in 1.1.5 works; I have not tried this, since this
       is not that important, and I don't really want to bother porting
       1.1.3 code from Sun or somewhere else.
    */
    ParkaApp (String  title)
    {
	super (title);

        WinV = new Vector();
	setLayout (new FlowLayout(FlowLayout.LEFT, 20, 10));
        add (new Button(NEWWIN));
        add (new Button(QUIT));
	pack();
	ParkaApp.apptype = new ApplType();
	ParkaApp.apptype.setApplication();
    }

    ParkaApp (String title, String[] args)
    {
	super(title);

	// If host, port and KB specified, load 'em up...
	if (args.length == 3)
	{
	    parkaHost = args[0];
	    try {
		parkaPort = Integer.parseInt(args[1]);
	    } catch (NumberFormatException ex) {
		parkaPort = 7990;
	    }
	    parkaKB = args[2];
	}

    }

    public boolean action (Event evt, Object arg)
    {
        if (evt.target instanceof Button)
        {
	    String label = (String) arg;
	    if (label.equals(NEWWIN))
  	    {
	        QueryWin win = new QueryWin (this, width, height,
					     parkaHost, parkaPort, parkaKB);
	        // win.setVisible (true); //Java 1.1 (replaces 'show()')
		        // Java on the SGI mach. is only 1.0.2
                win.show();
	        WinV.addElement (win);
	    }
	    else  if (label.equals(QUIT))
	    {
		// setVisible (false);   //Java 1.1 (replaces 'hide()')
		hide();
		dispose();
		System.exit(0);
	    }
	    return true;
        }
        else
	{
            return false;
	}
    }

    void set_url_for_icons()
    {
	FileInputStream  fis = null;
	DataInputStream  dis = null;

	try  {
	    fis = new FileInputStream (ICON_FILE);
	}
	catch  (FileNotFoundException fnf_e)  {
	    System.out.println ("Icon directory file: '" + ICON_FILE +
               "' expected in current directory was not found");
	    System.exit(1);
	}
	if (fis != null)
	{
	    try  {
		dis = new DataInputStream (fis);
		String  instr = null;
		url_directory = "";
		while ((instr = dis.readLine()) != null)
		{
		    url_directory = url_directory + instr;
		}
	    }
	    catch  (IOException io_e)   {
		System.out.println ("IO Error: " + io_e.getMessage());
	    }
	}
    }

    URL get_url_for_icons (String filename)
    {
	URL   urlloc = null;
	try  {
	    urlloc = new URL(url_directory + "/" + filename);
	}
	catch (MalformedURLException mal_e) {
	    System.out.println("Malformed url given: " + url_directory
                    + "/" + filename);
	    System.exit (1);
	}
	return  (urlloc);
    }

    public static void main (String[] args)
    {
        ParkaApp parkaAccess = null;

        if (args.length == 3)
	  parkaAccess = new ParkaApp("Parka Java Interface", args);
	else
	  parkaAccess = new ParkaApp("Parka Java Interface");

	// parkaAccess.setVisible (true);
	        // Java on the SGI mach. is only 1.0.2
	parkaAccess.show();
	parkaAccess.set_url_for_icons();
    }
}
