/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
package parka;

import java.awt.*;
import java.util.*;

/* This dialog is the general dialog class used for
   all the parka dialogs.
*/
/* NOTE:  Here at UMD, jdk 1.1.1 is the default version that is run.
   However, do not compile with it!   The dialogs sometimes SIGSEGV
   (segfault) with that version deep in the call to Dialog.show()
   somewhere-- a major threads bug, it looks like from all the stuff
   that prints out.  Maryland has jdk 1.1.5 now, tho it is not the
   default one to get.  Use that one on purpose.  The titles will not
   appear, but that is not a bug that will cause the application to
   totally fail, at least.
*/
public class ParkaDialog extends Dialog
{
    private ParkaWin parkawin;

    protected  int MIN_X = 300;
    protected  int MIN_Y = 200;

    // Button constants
    static final int CANCEL = 0;
    static final int OK = 1;

    static final String OKlabel = "Ok";
    static final String CANCELlabel = "Cancel";

    protected Button   ok = null,
                       cancel = null;

    /*******
     Methods
     *******/

    // get a modal dialog box with this constructor;
    // constructor below allows a choice
    //
    public ParkaDialog (ParkaWin parent, String title)
    {
        super (parent, title, true);

        parkawin = parent;
    }

    public ParkaDialog (ParkaWin parent, String title, boolean modality)
    {
        super (parent, title, modality);

        parkawin = parent;
    }

    public synchronized Dimension minimumSize()
    {
        return new Dimension (MIN_X, MIN_Y);
    }

    public synchronized Dimension preferredSize()
    {
        return minimumSize();
    }

    public boolean keyDown (Event e, int key)
    {
        if ((e.id == Event.KEY_PRESS) && (e.key == 10)) // 10 is return
        {
            answer(OK);
	    return  true;
        }
	else
            return false;
    }
  
    // Handle button events by calling the answer() method.
    // Pass the appropriate constant value, depending on the button.
    //
    public boolean action(Event e, Object arg)
    {
        if (e.target instanceof Button) 
        {
	    String buttonLabel = (String)arg;

            if (buttonLabel.equals(OKlabel)) 
                answer(OK);
            else  if (buttonLabel.equals(CANCELlabel))
                answer(CANCEL);
            else
                System.out.println ("No effect; No such button");

            return true;
        }
        else
            return false;
    }

    protected boolean list_action (Event e, Object arg, java.awt.List list)
    {
	if (e.target  instanceof  Button)
        {
	    String  buttonLabel = (String)arg;

	    if (buttonLabel.equals(OKlabel))
            {
                //assures its a list with one and only one item selected.
	        if (list.getSelectedIndex() != -1)
		{
                    answer(OK);
		    return   true;
		}
		else
                {
		    System.out.println ("Programmer Error:  " +
                             "List is not set up correctly");
                }
	    }
	    else  if (buttonLabel.equals(CANCELlabel))
	    {
                answer(CANCEL);
                return true;
	    }
            else
                System.out.println ("No effect; No such button");
        }
        return  false;
    }

    // Call ok(), or cancel() methods depending on the button the
    // user clicked.  Subclasses define how the answer is processed by
    // overriding this method or the  ok(), or cancel() methods.
    //
    protected void answer(int answer)
    {
        switch(answer)
        {
        case OK: 
            ok();
            break;
        case CANCEL: 
            cancel();
            break;
        }
        dispose();
    }

    // Subclasses should define Ok action
    protected void ok()
    {
        System.out.println ("Subclass did not define action");
    }
    
    // Do Cancel action
    protected void cancel()
    {
        parkawin.redrawGraphics();
    }

    /* Try to assure that redraws always occur when needed after dialogs close
       This has not been happening, and the drawing gets very disturbed until
       another Expose event occurs on the graphics canvas to make it redraw.
       With the next two routines we force it, and guarantee the redraw happens
       after the dialog is gone.
    */
    public void dispose()
    {
	super.dispose();
        parkawin.redrawGraphics();
    }

    public void hide()
    {
	super.hide();
        parkawin.redrawGraphics();
    }
}
