/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
package parka;

import java.net.*;
import java.util.*;
import java.io.*;
import java.awt.Frame;  //for Cursor
import java.awt.Window; //for dialogs

/**********************************************************************
  ParkaNet - networking interface for Parka
  **********************************************************************/
public class  ParkaNet  extends java.lang.Object
{
    // KB information
    protected  String KBName;
    protected  boolean usingMem;
             // 'useMem' refers to using memory vs. disk in internal
             // Parka operations, ie, joins.
    protected  String[] predicates;
    protected  Vector   predArity;

    protected  ParkaWin parent;

    protected  Socket       socket;
    protected  InputStream  instream;
    protected  OutputStream outstream;

    protected  NetKBInit  kbLoader;
    protected  boolean    kbOpen;
    protected  boolean    newPredList;

    // Hostname
    protected  String hostname;

    // Port number
    protected  int port;

    // Query results
    protected  int    numTotalAnswersToQuery;
           //number of total tokens from the server is going to be--
           //(if all of the answers were requested at once, rather than
           // in batches):
           //numTotalAnswersToQuery * getNumVariables() --
           //                  ie, num Rows * num Cols
    protected  int           varStatus;
    protected  String        numAnswersMsg;
    protected  QueryResults  qr;
/*  replace with  QueryResults.
    protected  String[]   answers;
*/
    protected  String[]   vars;

    static int   HAS_VARS = 1;
    static int   NO_VARS  = 2;
    static int   VARS_OTHER_STATUS  = 3;

    // Browser query results
    protected  String[][]  browserAnswers;
    protected  String      browserMsg;

    static final String  NOCONNMSG = "Not connected to server.";
    static final String  NOKBOPENMSG = "No KB open.";

    public String statusMsg;

    /*******
     Methods
     *******/

    public ParkaNet (ParkaWin parent)
    {
        // constructor
        this.parent = parent;
        socket = null;
	instream = null;
        outstream = null;

	initKBVars(true);
	initBrowserData();

	//default values:
	hostname = "theophilos.cs.umd.edu";
	port = 7990;

	statusMsg = null;
    }

    //some of these are also reinitialized with a new query
    private void initKBVars (boolean inConstruct)
    {
        setKBStatus (false);
        KBName = null;
	usingMem = true;   //may be reset later by chosen KB. Default.
        numTotalAnswersToQuery = 0;
	numAnswersMsg = null;
        predicates = null;
	predArity = null;
        qr = null;
        vars = null;
	varStatus = VARS_OTHER_STATUS;
	newPredList = true;   // 'predicates' not init'd yet

	if (!inConstruct)
            ((QueryWin)parent).updateFont();
    }

    void  initBrowserData()
    {
        browserAnswers = null;
	browserMsg = "";
    }

    void clearNumAnswersToQueryMsg()
    {
	numAnswersMsg = null;
    }

    void displayNumAnswersToQueryMsg()
    {
	if (numAnswersMsg != null)
	    parent.displayMessageKS (numAnswersMsg);
    }

    ParkaWin  getWin()
    {
	return   parent;
    }

    InputStream  getInStream()
    {
        return  instream;
    }

    OutputStream  getOutStream()
    {
        return  outstream;
    }

    public void connect()
    {
        // connect to Parka
        parent.displayMessageKS("Connecting to Parka");
		//'parent.displayMessage()' has unselect calls that are
		//redundant, when more msgs are bound to be posted.
        try 
        {
            // close socket if one exists
            if (this.isConnection())
                this.disconnect();

            // Create new socket and streams
            socket = new Socket (hostname, port);
            instream = socket.getInputStream();
            outstream = socket.getOutputStream();
            parent.displayMessageKS("Connection -- Waiting");
		// listening socket -- successful connect

	    is_connection_accepted();
        }
        catch (UnknownHostException uhe) 
        {
            parent.displayMessage("Connect Failed: Unknown Host Error");
        }
	/* BindException and ConnectException are java 1.1...
        catch (BindException be) 
        {
            parent.displayMessage("Connect Failed: " + be.getMessage());
	//TEST
	    System.out.println ("BindException");
        }
        catch (ConnectException ce) 
        {
            parent.displayMessage("Connect Failed: " + ce.getMessage());
	//TEST
            System.out.println ("Check if server is running " +
                           "or port is correct");
        }
	*/
        catch (SocketException se) 
        {
            parent.displayMessage("Connect Failed: Socket Failure");
	//TEST
	    System.out.println (se.getMessage());
        }
        catch (IOException ioe)
        {
            parent.displayMessage("Connect Failed: IO Error");
	//TEST
	    System.out.println (ioe.getMessage());
        }
	catch (IllegalArgumentException  iae) //possible with illegal port#
	{
            parent.displayMessage ("Connect Failed: " + iae.getMessage());
	//TEST
	    System.out.println (iae.getMessage());
	}

/* TEST
	System.out.println ("Leaving 'connect()'");
*/
    }

    private void is_connection_accepted()
    {
/* TEST
	System.out.println ("In is_connection_accepted()");
*/
        Response  r = new Response (this);
	r.computeResponse();
/* TEST
	System.out.println ("Response:  " + r.getResponse());
*/
	parent.displayMessage (r.getResponse());
    }

    public void disconnect()
    {
        // disconnect from Parka if connected
        if (isConnection())
        {
            // close open KB, if any
            if (isKBOpen())
	    {
		if (kbLoader.isThreadAlive())
		    kbLoader.join();
                closeKB();
	    }

            // now disconnect
            parent.displayMessage("Disconnecting from Parka");
            try
            {
                // Tell parka server we are exiting;
		// parka doesn't give a response
		Command c = new Command (this, Command.DISCONNECT, "");
                c.sendCommand();

                // Close streams
                instream.close();
                outstream.close();
                socket.close();
            
                instream = null;
                outstream = null;
                socket = null;
            }
            catch (IOException e1)
            {
                // Close Error
                System.out.println("IO Error in disconnect()");
            }
        }
    }

    // Returns true if there is a connection to PARKA open
    public boolean isConnection()
    {
        return ((socket != null) && (instream != null)
                            && (outstream != null));
    }

    // Change hostname and port used to connect to PARKA
    public void changeParkaHost(String newhost, int newport)
    {
        hostname = newhost;
        port = newport;
        
        // change PARKA connection
        this.disconnect();
        this.connect();
    }

    // Create PortDialog so user can change hostname and portnumber
    public void doPortDialog()
    {
        PortDialog pdialog = new PortDialog (parent, this, hostname, port);
	if (ApplType.getIsVersion1_0())
            ((Window)pdialog).show(); //Netscape thinks we want Dialog.show()
             //which doesn't exist until java 1.1
	else
            pdialog.show();
    }

    public synchronized boolean isKBOpen()
    {
	return   kbOpen;
    }

    public synchronized void setKBStatus (boolean status)
    {
	kbOpen = status;
    }

    // initialize connection to a new KB; uses a dialog box
    // eventually 'initLoadKBbyName()' gets called, once the KB
    // to load is known.
    //
    public void initKBLoading()
    {
	if (isConnection())  //assures the 2 streams already exist
	{
	    /* don't want to close existing KB (if any) and erase things,
               if the user decides after the dialog is up to cancel.  So
               this is commented out.
            if ((kbLoader != null) && kbLoader.isThreadAlive())
		kbLoader.join();
	    if (isKBOpen())
		closeKB();
            */
	    kbLoader = new NetKBInit (this, instream, outstream);
		//this class requires in/out streams to already be set up.
	    kbLoader.loadKB();
	}
	else
            parent.displayMessage (NOCONNMSG);
    }

    // initialize connection to a new KB without opening a dialog to 
    // request info from the user (Heflin, 5/21/98)
    //
    public void initLoadKBbyName (String newkbName, boolean memoryp)
    {
        /* added by Carolyn for multiple Parka server? */
        if ((kbLoader != null) && kbLoader.isThreadAlive())
	    kbLoader.join();
	if (isKBOpen())
	    closeKB();  //error not to close existing open KB first

	if (isConnection())  //assures the 2 streams already exist
	{
	    kbLoader = new NetKBInit (this, instream, outstream);
		//this class requires in/out streams to already be set up.
	    kbLoader.loadKBbyName (newkbName, memoryp);
	}
	else
            parent.displayMessage (NOCONNMSG);
    }

    // this routine is used from within a separate thread; so
    // we need to be careful about synchronization and interference.
    //
    public void openKB (String newkbName, boolean memoryp)
    {
	// System.out.println ("Entered 'openKB()'");

	if (isConnection())
        {
	    // System.out.println ("presence of connection is established.");

	    // not set from any other place:
	    KBName = newkbName;
	    usingMem = memoryp;
	    /* System.out.println ("   KBname = [" + KBName +
                        "] ; usingMem = " + usingMem);
            */
            // load a specific database
            parent.displayMessageKS("Loading " + KBName);
		//
		//don't call 'displayMessage()'; has unnecessary
		//unselect call overhead for this point in the code

            Command  c = new Command (this, Command.OPENKB, KBName);
	    c.sendCommand();
            Response r = new Response (this);
	    r.computeResponse();
	    // System.out.println ("   returned status is: " + r.getStatus());

            if (r.getStatus() == 0)
            { 
                setKBStatus (true);
            }

            parent.displayMessage (r.getResponse());
        }
    }

    public void closeKB()
    {
        if (isKBOpen())
        {
	    if (kbLoader.isThreadAlive())
	        kbLoader.join();

	    if (isConnection())
            {
                // Close the current KB
                parent.displayMessage("Closing KB");

                Command  c = new Command (this, Command.CLOSEKB, "");
                c.sendCommand();
                Response r = new Response (this);
	        r.computeResponse();

		if (r.getStatus() >= 0)  // succeeds
	        {
		    //reinit when closed
    		    initKBVars(false);
		    initBrowserData();
		    parent.reinitAppWin();
		}

                parent.displayMessage (r.getResponse());
            }
	    else
                parent.displayMessage (NOCONNMSG);
        }
	else
            parent.displayMessage ("No KB is open");
    } 

    // 'queryString' should be the standard Parka format: (pred#arg1#arg2...)
    // (pred#arg1#arg2...)...
    // Generates the entire QueryResults structure of answers, or null on error,
    // but does not return the answers itself.  In order to control large
    // numbers of answers, this is controlled through ResultsWin class.
    //
    public void query (String queryString)
    {
//	System.out.println("Entering query(\"" + queryString + "\")");
	statusMsg = null;            // Reset status

	if (!isConnection())
        {
	    statusMsg = NOCONNMSG;
	    return;
	}
	if (!isKBOpen())
	{
	    statusMsg = NOKBOPENMSG;
	    return;
	}

	qr = null;  //reinit

	if ((queryString == null) || (queryString.equals("")))
	{
	    statusMsg = "Empty query.";
	    return;
	}

	// Make sure that the KB has been loaded...
	if (kbLoader != null)
            kbLoader.join();

	Command c = new Command(this, Command.QUERY, queryString);
        c.sendCommand();

        Response r = new Response(this);
        r.computeResponse();

	numTotalAnswersToQuery = r.getStatus();
        if (numTotalAnswersToQuery < 0)       //then really have an error
        {
	    statusMsg = r.getResponse();
	    return;
        }
        else    //if (numTotalAnswersToQuery >= 0)
        {
	    String[] varList = simplyGetVariables();
/* TEST
	    System.out.println ("varList.length == " + varList.length);
	    for (int v = 0; v < varList.length; v++)
	        System.out.println ("   " + varList[v]);
*/
	    if (varStatus == NO_VARS)
	    {
                numAnswersMsg = " Answer is TRUE";
                statusMsg = "Done. " + numAnswersMsg;
	        return;  // constant expression found
	    }
            else   //varStatus == HAS_VARS
	    {
                numAnswersMsg = " " + numTotalAnswersToQuery
                                       + " answers.";
                statusMsg = numAnswersMsg;

	        qr = new QueryResults (numTotalAnswersToQuery, varList.length);
	        qr.setNames (varList);
                qr.setValueMatrix (simplyGetAnswers());
	    }
	}
    }

    public String[] simplyGetVariables()
    {
	if (!isConnection())
        {
	    statusMsg = NOCONNMSG;
	    return null;
	}
	if (!isKBOpen())
	{
	    statusMsg = NOKBOPENMSG;
	    return null;
	}

	Command  c = new Command (this, Command.GETVARIABLES, "");
	c.sendCommand();
	Response r = new Response (this);
	r.computeResponse();

	String[]  varlist = null;

	if (r.getStatus() > 0)
	{
	    varStatus = HAS_VARS;
	    r.makeTokens();
	    varlist = r.getTokenList();
	}
        else  if ((getNumTotalAnswersToQuery() == 0) &&
                       (r.getResponse()).equals(ResultsWin.NOVARSINQUERY))
                                 // [r.getStatus() == -1, in this case]
	{
	    varStatus = NO_VARS;
	}
	else
	{
	    varStatus = VARS_OTHER_STATUS;
	}
	return  varlist;
    }

    String[][] simplyGetAnswers()
    {
	int numResults = qr.numResults();
	int numNames = qr.numNames();

	String[] ansList;
	String[][] ansMat = new String[numResults][numNames];

	if (!isConnection())
        {
	    return null;
	}

	Command  c = new Command (this, Command.GETBINDINGS,
				  1 + " " + numResults * numNames);
	c.sendCommand();
	Response r = new Response (this);
	r.computeResponse();

	if (r.getStatus() <= 0)
	{
	    System.out.print ("ParkaNet.simplyGetAnswers():  ");
	    System.out.println (" Bad status from server.");
	    return null;
	}

	r.makeTokens();
	ansList = r.getTokenList();
	if (ansList.length != numResults * numNames)
	{
/* TEST
	    System.out.println ("ParkaNet.simplyGetAnswers():  ");
	    System.out.println ("Size of results doesn't match number of tokens.");
*/
	    /* It turns out that the most likely reason for this is that the
               client asked for a chunk of data too large to fit into the
               server's buffer.  If the server's buffer fills up, it just
               quits, and tells us how many rows it was able to send, out of
               the number we asked for.  So we either need to accept what we
               have, or ask for the rest;  we are going to keep asking until
               we get the total number we wanted.
            */

	    if (r.getStatus() < 0)  //error
		return  null;
/* TEST
	    System.out.println ("   Answer list length:  " + ansList.length);
	    System.out.println ("   Answer list length:  " + ansList.length);
	    System.out.println ("   Number of rows:  " + numResults);
	    System.out.println ("   Number of columns:  " + numNames);
*/
	    int  numGotten = r.getStatus();
	    int  leftToGet = numResults - numGotten;

	    // Convert answer list into value matrix...
	    makeValueMatrix (0, numGotten, numNames, ansList, ansMat);

	    while (leftToGet > 0)
	    {
        	c = new Command (this, Command.GETBINDINGS,
                                (numGotten +1) + " " + leftToGet * numNames);
			/* we'll ask for this much; may not get it all ... */
        	c.sendCommand();
        	r = new Response (this);
        	r.computeResponse();
		if (r.getStatus() > 0)
	            r.makeTokens();
		else
		{
		    System.out.println ("Unexpected error retreiving answers.");
		    System.out.println ("   Status == " + r.getStatus());
		    return  null;
		}
        	ansList = r.getTokenList();

	        makeValueMatrix (numGotten, r.getStatus(),
                                    numNames, ansList, ansMat);

		leftToGet -= r.getStatus();
		numGotten += r.getStatus();
	    }
	}
	else {
	  // Convert answer list into value matrix...
	  makeValueMatrix (0, numResults, numNames, ansList, ansMat);
	}

	return ansMat;
    }

    private void makeValueMatrix (int rowStart, int numRows, int numCols,
                                  String[] ansList, String[][] ansMat)
    {
/* TEST
	System.out.println ("VALUE MATRIX:  rowStart == " + rowStart +
                     " ; last row == " + (rowStart + numRows -1) +
                           " ; numCols == " + numCols);
*/ 
	for (int k = 0, i = rowStart; i < rowStart + numRows; i++, k++)
	{
	    for (int j = 0; j < numCols; j++)
	    {
		ansMat[i][j] = ansList[k * numCols + j];
	    }
	}
    }

/////////////////// OLD METHODS ////////////////////

    // 'links' are the predicates in the query graph
    public void submitQuery (Vector links)
    {
	if (!isConnection())
        {
	    parent.displayMessage(NOCONNMSG);
	    return;
	}
	if (!isKBOpen())
	{
	    parent.displayMessage("No open KB");
	    return;
	}

        kbLoader.join();
        parent.displayMessage("Submitting Query");

        // submit a query to parka engine
        StringBuffer querystring = new StringBuffer();

        /*  gather query info */
        Enumeration preds = links.elements();
        while (preds.hasMoreElements())
        {
            Link currlink = (Link)(preds.nextElement());
            querystring.append ("(" + currlink.getObjName());

	    Vector linkArgs = currlink.getLinkArguments();

	    for (int i = 0; i < linkArgs.size(); i++)
	    {
                querystring.append ("#");
		querystring.append ((String) linkArgs.elementAt(i));
	    }
            querystring.append (")");
        }

        // error checking
	String   finalQuery = new String (querystring);
	if (finalQuery.equals(""))
	{
            parent.displayMessage ("No query formed");
	    return;
	}

        Command  c = new Command (this, Command.QUERY, finalQuery);
        c.sendCommand();
        Response r = new Response (this);
        r.computeResponse();

	numTotalAnswersToQuery = r.getStatus();

        if (numTotalAnswersToQuery < 0)   //then really have an error
        {
/*  TEST --
    what prints if we let 'r.getResponse()' print??

	    if (numTotalAnswersToQuery != Response.XCEPTION)
	    {
		// either boolean answer of "False", or no way to match
		// the query graph to any data in the database.  Could
		// include use of constants not in the database, or in
		// the database in a different context (eg. a predicate
		// used in a domain or range position)
	        //
                parent.displayMessage ("No results found");
	    }
	    else
*/
	        parent.displayMessage (r.getResponse());

            qr = null;
            vars = null;
        }
        else //if (numTotalAnswersToQuery >= 0) 
        {
            vars = getVariables();

	    if (varStatus == NO_VARS)
	    {
		numAnswersMsg = " Answer is TRUE";
                parent.displayMessage ("Done. " + numAnswersMsg);
	    }
	    else   //varStatus == HAS_VARS
	    {
		numAnswersMsg = " " + numTotalAnswersToQuery
                                       + " answers.";
                parent.displayMessage ("Done. " + numAnswersMsg);
	    }
        }
    }

    String getBrowserMsg()
    {
	return  browserMsg;
    }

    private void printBrowserSampleMsgs (Vector coltype, Vector numans,
                                         Vector numshown)
    {
/*  No browser in this version

	int  vectsize = coltype.size();
	if (vectsize != numans.size())
	{
	    System.out.println ("Programmer Error: column type vector "
                 + "and number of answers in each column vector are "
                 + "not the same size.");
	    return;
	}

	StringBuffer msgBuf = new StringBuffer (
                     "Too much data; only a sample is used.");

	for (int i = 0; i < vectsize; i++)
	{
	    if ((i % 2) == 1)
		msgBuf.append ("\nColumn ");
	    else
	        msgBuf.append ("  Column ");
	    String  col = BrowserWin.getColumnName (
                                               (String)coltype.elementAt(i));
	    String  num = (String)numans.elementAt(i);
	    String  shown = (String)numshown.elementAt(i);
	    msgBuf.append (col);
	    msgBuf.append (" has ");
	    msgBuf.append (num);
	    msgBuf.append (" total answers. ");
	    msgBuf.append (shown);
	    msgBuf.append (" shown.");
	}
	browserMsg = new String (msgBuf);
*/
    }

    // Browsing query has a specialized expected structure --
    // See comments on server side.
    // Returns contents of an error message, if any.
    //
    int  submitBrowsingQuery (String  query)
    {
	int  status;
	browserMsg = "";
	status = -4;
	System.out.println ("Browser not available for this version");

/*      Exclude from Nary Client:

	browserMsg = "";
        Command  c = new Command (this, Command.JBROWSERQUERY, query);
        c.sendCommand();
        BrowserResponse r = new BrowserResponse (this);
        r.computeResponse();
	status = r.getStatus();

	if (status >= -1)   // 0 is the highest possible
	{
	    r.makeTokens();
	    browserAnswers = r.getBrowserTokenList();
	    if (status == -1)   // sample of all the data only
	    {
	        Vector  columnType = r.getColType();
	        Vector  numberAns = r.getNumAnswers();
		Vector  numberShown = r.getNumAnswersShown();
	        printBrowserSampleMsgs (columnType, numberAns, numberShown);
	    }
	}
	if (status == -2)  //corrs to JB_INCOMPLETEDATA; get more data:
	{
	    r.makeTokens();

	    while (status == -2)  // 0 is a successful end
	    {
		c = new Command (this, Command.JBRWQUERYCONT, "");
		c.sendCommand();
		r.computeResponse();
		status = r.getStatus();

		if (status >= -2)
		{
	            r.makeMoreTokens();
		}
	    }
	    if (status >= -1)
	    {
	        browserAnswers = r.getBrowserTokenList();
	        if (status == -1)
		{
	            Vector  columnType = r.getColType();
	            Vector  numberAns = r.getNumAnswers();
		    Vector  numberShown = r.getNumAnswersShown();
	            printBrowserSampleMsgs (columnType, numberAns, numberShown);
		}
	    }
	    else
	        browserMsg = r.getResponse();
	}
	else  if (status == -3)
	    browserMsg = "No more data received; server out of memory";
	else  if (status == -4)
	    browserMsg = r.getResponse();
*/
	return  status;
    }

    // Set up query results in the query results object, ultimately
    // for display.
    // Should not call this routine if no query has previously
    // been sent.
    // get start..end answers from the server.
    // start must be > 0; end must be > start.
    //
    void getAnswersForQResults (int start, int end, QueryResults qr)
    {
	if (isConnection())
        {
	    //Nothing to process here:
	    if (numTotalAnswersToQuery <= 0)
		return;

	    if ((start < 1) || (end < 1))
            {
		System.out.println ("ParkaNet.getAnswersForQResults(): "
                   + "'start' 'and' end parameters must be greater than zero");
		return;
            }
	    if (end < start)
            {
		System.out.println ("ParkaNet.getAnswersForQResults(): "
		      + "'end' parameter cannot be less than 'start'");
		return;
            }

            Command  c = new Command (this, Command.GETBINDINGS,
                                   start + " " + end);
            c.sendCommand();
            Response r = new Response (this);
            r.computeResponse();
	    r.makeTokens();
	    String [] answers = r.getTokenList();
	    qr.setValueMatrix (answers);

	    /* TEST
	    if (answers == null)
	        System.out.println ("'answers' is null");
	    else
	        System.out.println ("'answers' has " + answers.length
                                             + " answers.");
	    */
	    // this will obtain tokens for row x column responses
            // num variables is number of columns; number of rows
            // is total/columns.
        }
        else
	{
            parent.displayMessage (NOCONNMSG);
	}
    }

    public QueryResults getQueryResults()
    {
        return   qr;
    }

    public String[][] getBrowserAnswers()
    {
        return   browserAnswers;
    }

    public String[] getVariables()
    {
	if (isConnection())
        {
            Command  c = new Command (this, Command.GETVARIABLES, "");
	    c.sendCommand();
            Response r = new Response (this);
	    r.computeResponse();

	    if (r.getStatus() > 0)
	    {
	        r.makeTokens();
	        vars = r.getTokenList();
	        varStatus = HAS_VARS;
	    }
	    else  if ( (getNumTotalAnswersToQuery() == 0) &&
                       ((r.getResponse()).equals(ResultsWin.NOVARSINQUERY)) )
                                 // [r.getStatus() == -1, in this case]
	    {
	        varStatus = NO_VARS;
		vars = null;
	    }
	    else
	    {
	        varStatus = VARS_OTHER_STATUS;
		vars = null;
	    }
        }
	else
        {
            parent.displayMessage (NOCONNMSG);
            vars = null;
        }

        return vars;
    }

// Not currently used:
    public void getStats()
    {
        // Get the size stats for the currently loaded KB
	if (isConnection())
        {
            Command  c = new Command (this, Command.PARKACONFIG, "");
	    c.sendCommand();
            Response r = new Response (this);
	    r.computeResponse();
	    // no display done
        }
	else
	    parent.displayMessage (NOCONNMSG);
    }

// Not currently used:
    public void getTime()
    {
        // Get the time information for the last query
	if (isConnection())
        {
            Command  c = new Command (this, Command.GETTIMINGS, "");
	    c.sendCommand();
            Response r = new Response (this);
	    r.computeResponse();

	    /* added for statistics */
	    String[] stats;
	    int i;

	    r.makeTokens();
	    System.out.println("Server timings:");
	    stats = r.getTokenList();
	    for (i=0; i < stats.length; i++)
	      System.out.println(stats[i]);

	    // no display done
        }
	else
	    parent.displayMessage (NOCONNMSG);
    }

    // Gets a list of predicates relevant to a particular database.
    // User class is expected to call this routine after the thread
    // loading the requested KB has successfully completed.
    //
    private String[] getPreds()
    {
	String[]  predList = null;

        // Get predicate list from PARKA
	if (isConnection())
        {
	    if (isKBOpen())
	    {
	        if (!kbLoader.isThreadAlive())
	        {
		// using the "KS" version of displayMessage so that the
		// active Node is not disturbed (When the link/predicate
		// is created, we need to know this to assign the "from"
		// node. (KS == keep state).
                parent.displayMessageKS("Retrieving Predicates");

		boolean  morePredsToGet = true;
                boolean  inLoopAgain = false;
                while (morePredsToGet)
		{
		    String[] tmp, tmp2;

                    Command  c = new Command (this, Command.GETPREDICATES, "");
	            c.sendCommand();
                    Response r = new Response (this);
	            r.computeResponse();

                    // status == 0 if have no more preds to get.
		    if (r.getStatus() >= 0)
		    {
			// System.out.println ("get preds:  status > or = 0");
	                r.makeTokens();
	                tmp = r.getTokenList();
		    }
		    else //r.getStatus() < 0: error occurred while getting preds
		    {
			// System.out.println ("get preds:  status < 0");
		        predList = null;
		        parent.displayMessage (r.getResponse());
			break;
		    }

		    morePredsToGet = (r.getStatus() > 0);
		    if (inLoopAgain)
		    {
			// System.out.println ("  get preds:  " +
                        //              "Getting preds mult times thru loop");
			tmp2 = predList;
			predList = new String[tmp.length + tmp2.length];
                        // keep in original order:
			System.arraycopy (tmp2, 0,
			          predList, 0, tmp2.length);
			System.arraycopy (tmp, 0,
			          predList, tmp2.length, tmp.length);
		    }
		    else
			predList = tmp;

		    inLoopAgain = true;
		}
                // if r.getStatus() == 0, we have all the preds, and
                // everything is ok.
	        }
	        else
		    parent.displayMessageKS("Predicates not ready yet");
	    }
	    else
	    {
		parent.displayMessage("No KB open");
	        predList = null;  //assure null value when no connection
	    }
        }
	else
	{
	    parent.displayMessage (ParkaNet.NOCONNMSG);
	    predList = null;  //assure null value when no connection
	}
        /* TEST
        if (predList == null)
            System.out.println ("Predicates are null");
        else
            System.out.println (predList.length + " Predicates");
        */
        return    predList;
    }

    // should not be called if server call GETPREDICATES failed.
    //
    private Vector getPredsArity()
    {
	Vector pArity = new Vector();

	Command c = new Command (this, Command.GETPREDARITYS, "");
	c.sendCommand();
	// System.out.println ("In 'getPredsArity()'");

	Response r = new Response (this);
	r.computeResponse();

	if (r.getStatus() == 0)
	{
	    r.makeTokens();
	    String[]  arityList = r.getTokenList();
	    for (int i = 0; i < arityList.length; i++)
		pArity.addElement (arityList[i]);
	}
	else
	    parent.displayMessage (r.getResponse());

	return  pArity;
    }

    /* When we know that valid commands have been responded to, without
       error for the commands GETPREDICATES and GETVARIABLES, then it
       is valid to call the next two routines.    */

    // Returns the predicate array
    // used 'getPreds()' when KB was opened initially, to retrieve
    // the predicates from Parka
    //
    public String[] getPredicateList()
    {
	if (isKBOpen() && newPredList)
        {
	    if (kbLoader.isThreadAlive())
	        kbLoader.join();

	    predicates = getPreds();
	    newPredList = false;  //if predicates is still null,
                   // something is wrong.

	    if (predicates != null)
	    {
	        predArity = getPredsArity();
		     // need arity for each predicate, but client's user
		     // doesn't need to know this call is needed.
		     // its done 'automatically' from user's view.

	        if (predicates.length != predArity.size())
		    System.out.println ("Programming error: list of predicates"
                       + " and list of pred arities are not equal length");
	    }
        }
	else  if (!isKBOpen())
	    parent.displayMessage ("No KB is open");

        return predicates;
    }

    // should return 'null' if 'getPredicateList()' has not been called first.
    //
    public int getPredicateArity (String predname)
    {
	if (predicates == null)
	    return  0;

	int i;
	for (i = 0; i < predicates.length; i++)
	    if (predname.equals (predicates[i]))
		break;

	if (i > predArity.size())
	    return  0;
	else
	    return  Integer.parseInt ((String)predArity.elementAt(i));
    }

    public String[] getVariableList()
    {
        return vars;
    }

    //same as number of columns for output display of query results
    //
    public int getNumVariables()
    {
	if (vars == null)
	    return   0;
	else
	    return   vars.length;
    }

    //total number of answers generated at the server
    //
    public int getNumTotalAnswersToQuery()
    {
	return  numTotalAnswersToQuery;
    }
}
