/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
package parka;

import java.awt.*;

public class PredDialog extends ParkaDialog
{
    DrawArea  drawArea;
    boolean   havelist;

    // Button constants
    public static final int CANCEL = 0;
    public static final int OK = 1;

    static final String  EMPTYLIST = "Empty predicate list";
    protected static final String TITLE = "Please select a predicate";
    //protected static final int NUM_ROWS = 5;

    protected List list;

    public PredDialog (ParkaWin parentwin, DrawArea da, String[] preds)
    {
        // Create the window.
        super (parentwin, TITLE);

        drawArea = da;

        // Specify a LayoutManager for it
        setLayout(new BorderLayout(15, 15));

        // Set up list
        list = new List(10, false);
        if (preds == null)
	{
            list.addItem (EMPTYLIST);
	    havelist = false;
	}
        else
        {
	    for (int i = 0; i < preds.length; i++)
                  list.addItem (preds[i]);
            list.select(0);  // preselect first item
	    havelist = true;
        }

        // Put the list
        this.add("Center", list);

        // Create a panel of buttons, center the row of buttons in
        // the panel, and put the pane at the bottom of the window.
        Panel p = new Panel();
        p.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 15));
	if (havelist)
            p.add(ok = new Button (OKlabel));
        p.add(cancel = new Button (CANCELlabel));
        this.add("South", p);

        // Set the window to its preferred size.
        this.pack();
    }

    public synchronized Dimension preferredSize()
    {
	return  (new Dimension (MIN_X, MIN_Y + 100));
    }
  
    // Handle button events by calling the answer() method.
    // Pass the appropriate constant value, depending on the button.
    public boolean action (Event e, Object arg)
    {
	return  (super.list_action (e, arg, list));
    }

    // Do Ok action
    protected void ok()
    {
        String pred = list.getSelectedItem();
/* TEST
	System.out.println("Pred is: [" + pred + "]");
*/
        if (pred == null)
            cancel();
        else
	    drawArea.doLinkAction (pred);
    }

    protected void cancel()
    {
	drawArea.cancelEdge();
	super.cancel();
    }
}
