/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
package parka;

import java.util.*;
import java.io.*;

/**********************************************************************
  Response - a response object from parka
  **********************************************************************/
public class Response extends java.lang.Object
{
    protected  int      status;
    protected  int      size;
    protected  boolean  reqsTokeniz;
    protected  String   response;
    protected  InputStream instream;

	 //for delimiting tokens.
    protected  static final  String  DELIM = "\t";
         // The Tokenizer, used in makeTokens() in this class,
         //   applies to some commands.
         // Examples for Parka are:  GETVARIABLES, GETPREDICATES,
         //   GETKBNAMES, and GETBINDINGS
    protected  String[]  tokenList;
      // we need the following values also, with GETBINDINGS.
    protected  int numTokens;
      // for Query responses:
    protected  int numResponses;    //same as size, when successful query

    final static int     XCEPTION = -10;
	//when this kind of error happens, might want to use the
	//'response' field to display to ultimate user.

    /*  ************  */
 
    public Response (ParkaNet pn)
    {
	size = 0;
        status = -1;
	response = "";
	// init query related member variables:
	tokenList = null;
	numTokens = 0;
	numResponses = 0;    //same as size, when successful query
	instream = pn.getInStream();  //input stream assoc. with socket
    }

    public int getStatus()
    {
	return  status;
    }

    public int getSize()
    {
	return  size;
    }

    public String getResponse()
    {
	return  response;
    }

    // Compute PARKA's response:  status + " " + length + " " + response
    /* The format follows the parka server format for responses to
       requests, which is:
           1. status number, blank space
           2. size of the response (# chars, not including terminating '\0')
                 [server is in C code], blank space
	   3. the response

       In the case where we are getting our first response after submitting
       a query, the 1st number back is the status if it is less than zero,
       BUT, it is the number of answers if it is greater than or equal to 0.

       The 2nd number is the string length of the response, as above, and the
       3rd number is the response.

       A fuller explanation of the responses from the server is in the
       directory with the server code and description.
    */
    void computeResponse()
    {
        try
        {
            StringBuffer in = new StringBuffer();
            int tmp;
            boolean exception_occured = false;

            // Get status;  read until first space
            tmp = instream.read();
            while ((tmp != -1) && (((char)tmp) != ' '))
            {
                in.append((char)tmp);
                tmp = instream.read();
            }

            try
            {
                status = Integer.parseInt(in.toString());
            }
            catch (NumberFormatException e)
            {
                System.out.println("Error parsing status: [" +
                                        in.toString() + "]");
		exception_occured = true;
            }

	    // Get size of incoming response
            in = new StringBuffer();

            tmp = instream.read();
            while ((tmp != -1) && (((char)tmp) != ' '))
            {
                in.append((char)tmp);
                tmp = instream.read();
            }

            try
            {
                size = Integer.parseInt(in.toString());
            }
            catch (NumberFormatException e)
            {
                System.out.println("Error parsing size: [" + in.toString()
                                          + "]");
		exception_occured = true;
            }

	    if (!exception_occured)
	    {
                // Get response
                byte[] bytearr = new byte[size];
                int readsize = 0;
                //System.out.println("Reading");
                while (readsize < size)
                {
                    readsize += instream.read (bytearr, readsize,
                                                (size-readsize));
                }
                response = new String(bytearr, 0);
            }
	    else
	    {
		response = "Communication failure with Server";
		size = response.length();
		status = XCEPTION;
	    }

            // output will be:
/* TEST
            System.out.println ("status = "
                                 + status + "; size = " + size +
                                 "; response = " + response);
*/
        }
        catch (IOException e1)
        {
            System.out.println ("IO Error in getResponse()");
	    status  = XCEPTION;
	    response = "Communication failure with Server";
	    size    =  response.length();
        }
    }

    // if the Response requires making tokens out of the String, this
    // routine may be called.
    //
    protected void makeTokens()
    {
	int i = 0;
	StringTokenizer  responseTokens = new StringTokenizer (response, DELIM);
	numTokens = responseTokens.countTokens();
        tokenList = new String[numTokens];
/* TEST
        System.out.println ("Token list (answer) is " + numTokens + " long");
*/
	while (responseTokens.hasMoreTokens())
	{
	    tokenList[i] = responseTokens.nextToken();
	    /*
	      System.out.println ("Token: " + tokenList[i]);
	    */
	    i++;
	}
    }

    public String[] getTokenList()
    {
	return   tokenList;
    }
}
