/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */

package  parka;

import   java.awt.*;
import   java.util.*;
import   java.applet.*;
import   java.net.*;

/* For managing displayed results of a database query
 */
/* NOTE: This class was called 'QueryResults' in older versions;
   QueryResults is now a different class
 */
public class ResultsWin  extends Panel  implements MultiColumnCallback
{
    QueryWin      parentWin;
    MultiColumn   resultList;
    Thread        resultThread;
    ParkaNet      pn;
    DrawArea      da;

    static private String   START_LBL_STR = "First Item: ";
    static private String   END_LBL_STR = "Last Item: ";
    static private String   NEXT_CB_STR = "Get Next Group";
    static private String   PREV_CB_STR = "Get Last Group";
    static private int      NUMCOLS_TEXTF = 10;
    static private int      GRIDROWS = 2;
    static private int      GRIDCOLS = 2;
    static private int      HORGAP = 2;
    static private int      VERTGAP = 2;

    Panel         outputIndicators;
    TextField     startText;
    TextField     endText;
    Checkbox      nextCB;
    Checkbox      prevCB;
    CheckboxGroup cg;

    int           totalNumAnswers;
    int           numCols;
    int           firstMultiListItem;
    int           lastMultiListItem;
    static private int    blockSize; //number of answers obtained per call
                                     //to server
    private int   selectedRow = -1;   // number of result row last selected

    AppletContext context;    // used to load documents in web-browser

    static final String   NOVARSINQUERY = "No variables retrieved";
	//MUST match value returned from server when request for
	//variables is given.

    ResultsWin(QueryWin parWin, ParkaNet netConnect, DrawArea querySrc)
    {
        super();

        parentWin = parWin;
        resultList = null;
        resultThread = null;
	pn = netConnect;
	da = querySrc;

/* TEST -- make very small, for test.
	blockSize = 5;
	System.out.println ("blockSize == " + blockSize);
*/
	blockSize = 100;      //arbitrary choice, but small enough not
                              //to upset Vectors inside MultiColumn.

        setupIndicatorPanel();
	initIndicators(); //do after 'setupIndicatorPanel()'
        setLayout (new BorderLayout (0, 0));

        //create a single empty MultiColumn, to start.
	emptyMultiColumn();
        validate();

        /* Heflin 5/9: Code for applet context. Used to load html documents */
	Container grandparent;

	/* determine if this is in an Applet */
	grandparent = parentWin.parent;
	if (grandparent instanceof Applet) 
	  context = ((Applet)grandparent).getAppletContext();
	else
	  context = null;
    }

    //create a single empty MultiColumn
    void emptyMultiColumn()
    {
        String[] initlist = new String[1];
        initlist[0] = " ";
        resultList = new MultiColumn (initlist);
		//single empty column header
        add("North", outputIndicators);
        add("Center", resultList);
        validate();
    }

    void initIndicators()
    {
	firstMultiListItem = 0;
	lastMultiListItem = 0;
	numCols = 0;
	totalNumAnswers = 0;
	setTextFields();
        prevCB.disable();  //must have already created.
        nextCB.disable();  //must have already created.
	pn.clearNumAnswersToQueryMsg();
    }

    private void setupIndicatorPanel()
    {
	outputIndicators = new Panel();
	outputIndicators.setLayout (new GridLayout(GRIDROWS, GRIDCOLS,
                        HORGAP, VERTGAP));
	Label  start = new Label (START_LBL_STR, Label.LEFT);
	Label  end = new Label (END_LBL_STR, Label.LEFT);
	startText = new TextField (NUMCOLS_TEXTF);
	startText.setEditable (false);
	endText = new TextField (NUMCOLS_TEXTF);
	endText.setEditable (false);
	Panel  txtIndStart = new Panel();  //FlowLayout is default
	txtIndStart.add (start);
	txtIndStart.add (startText);
	Panel  txtIndEnd = new Panel();    //FlowLayout is default
	txtIndEnd.add (end);
	txtIndEnd.add (endText);

	nextCB = new Checkbox (NEXT_CB_STR);
	prevCB = new Checkbox (PREV_CB_STR);
	cg = new CheckboxGroup();
	prevCB.setCheckboxGroup(cg);
	nextCB.setCheckboxGroup(cg);
	cg.setCurrent (nextCB);

        outputIndicators.add (txtIndStart);
        outputIndicators.add (txtIndEnd);
        outputIndicators.add (prevCB);
        outputIndicators.add (nextCB);
    }

    private void setTextFields()
    {
	startText.setText (String.valueOf(firstMultiListItem));
	endText.setText (String.valueOf(lastMultiListItem));
    }

    void clear()
    {
	initIndicators();
	if (resultList != null)
	{
	    resultList.clear();
	    resultList.erase_titles();
	}
    }

    public boolean action (Event e, Object ob)
    {
	if (e.target  instanceof  Checkbox)
	{
	    Checkbox cb = (Checkbox)e.target;
	    if ( (cb.getLabel()).equals (NEXT_CB_STR))
	    {
		firstMultiListItem += blockSize;
		lastMultiListItem += blockSize;
		if (lastMultiListItem >= totalNumAnswers)
		{
		    lastMultiListItem = totalNumAnswers;
		    nextCB.disable();
		}
		setTextFields ();
		if (!prevCB.isEnabled())
		    prevCB.enable();
		putAnswersInMultiList (pn.getQueryResults());
		return   true;
	    }
	    else  if ( (cb.getLabel()).equals (PREV_CB_STR))
	    {
                lastMultiListItem = firstMultiListItem - 1;
                firstMultiListItem = (lastMultiListItem - blockSize) + 1;

		if (firstMultiListItem <= 1)
		{
		    firstMultiListItem = 1;
		    prevCB.disable();
		}
                //readjust, to make sure we have a whole blocksize of answers:
		if (firstMultiListItem == 1)
		    lastMultiListItem = blockSize;

		setTextFields ();
		if (!nextCB.isEnabled())
		    nextCB.enable();
		putAnswersInMultiList (pn.getQueryResults());
		return   true;
	    }
	    else
	        System.out.println ("Programmer Error -- " +
                          "Cannot recognize checkbox used");
	}
	return   false;
    }


    // 'links' are the predicates in the query graph
    private String createQueryString(Vector links)
    {
        // submit a query to parka engine
        StringBuffer querystring = new StringBuffer();

        //  gather query info
        Enumeration preds = links.elements();
        while (preds.hasMoreElements())
        {
	    Link link = (Link) preds.nextElement();
	    querystring.append(createAssertionString(link.getObjLabel(),
						     link.getLinkArguments()));
        }

	return new String (querystring);
    }

    private String createAssertionString(String predicate, Vector args)
    {
	if ((predicate == null) || (args == null))
	{
	    return null;
	}

        StringBuffer buffer = new StringBuffer();

	buffer.append ("(" + predicate);
	for (int i = 0; i < args.size(); i++)
	{
	    buffer.append ("#");
	    buffer.append (escapeSpecialChars((String) args.elementAt(i)));
	}
	buffer.append (")");

	return buffer.toString();
    }

    private  void initAnswersForMultiList (QueryResults  qr)
    {
	totalNumAnswers = qr.numResults();
	if (totalNumAnswers > 0)
            numCols = qr.numNames();
	else
	    numCols = 0;

	if (prevCB.isEnabled())
	    prevCB.disable();

	if (totalNumAnswers > blockSize)
	    nextCB.enable();
	else if (nextCB.isEnabled())
	    nextCB.disable();

	// call for first set of answers for the user.
	firstMultiListItem = 1;
        lastMultiListItem = Math.min (blockSize, totalNumAnswers);
	setTextFields();
	putAnswersInMultiList (qr);
        validate();
    }

    //Must be called sometime after pn.submitQuery
    //and after firstMultiListItem and lastMultiListItem
    //are correctly set.
    //
    private  void putAnswersInMultiList (QueryResults qr)
    {
	// If we do not call 'MultiColumn.clear()', answers just
	// accumulate.  Of course that would require different handling
	// than we do here, since when a user would go back and forward
	// multiple times with the current way of doing things, we'd
	// just accumulate more and more answers (same ones appending).
	//
	resultList.clear();

	int numRows = Math.min (blockSize, (lastMultiListItem -
                                   firstMultiListItem) + 1);
	int  start = firstMultiListItem -1;
	int  end   = start + numRows;
/* TEST
	System.out.println ("numRows == " + numRows);
	System.out.println ("start == " + start);
	System.out.println ("end == " + end);
*/
	for (int i = start; i < end; i++)
	{
	    resultList.addItem(qr.getValues(i));
	}
	validate();
	parentWin.validate();
/* TEST
	System.out.println ("Full QueryResults are:\n");
	System.out.println (qr.toString());
*/
    }

    // doQuery submits the query to Parka.
    public void doQuery()
    {
	resultList = null;
	removeAll();            //delete answers from an old query

	String queryString = createQueryString(Link.linksVector);
	pn.query (queryString);
	QueryResults qr = pn.getQueryResults();

	if (qr == null)
	{
	    parentWin.displayMessage (pn.statusMsg);
	    initIndicators();
	    emptyMultiColumn();
            validate();
	}
        else
        {
	    // Remove duplicate entries, since they are unnecessary...
	    qr.removeDuplicates();

	    if (qr.numResults() == 1)
	    {
		parentWin.displayMessage("1 answer found.");
	    }
	    else
	    {
		parentWin.displayMessage (qr.numResults() + " answers found.");
	    }

	    resultList = new MultiColumn (qr.getNames(), this);
	    
            add ("North", outputIndicators);
            add ("Center", resultList);
            validate();

            initAnswersForMultiList (qr);
        }
    }

  /*
    // doQuery submits the query to Parka.
    public QueryResults doQuery()
    {
      // added for timing statistics 
        Date startTime, endTime;
	startTime = new Date();

	resultList = null;
	removeAll();            //delete answers from an old query

	String queryString = createQueryString(Link.linksVector);
	if (queryString == null)
	{
	    parentWin.displayMessage ("No query formed");
	    return null;
	}

	QueryResults qr = pn.query(queryString);
	if (qr == null)
	{
	    parentWin.displayMessage("No results found in KB.");
	    initIndicators();
	    emptyMultiColumn();
            validate();
	    return null;
	}
	else if (qr.numResults() == 0)
	{
	    if (qr.existsInKB() == true)     // constant query in KB
	    {
		parentWin.displayMessage ("Answer: TRUE");
	    }
	    else
	    {
		parentWin.displayMessage ("Answer: FALSE");
	    }
	    initIndicators();
            validate();
	    return null;
	}
        else
        {
	    // Remove duplicate entries, since they are unnecessary...
	    qr.removeDuplicates();

	    if (qr.numResults() == 1)
	    {
		parentWin.displayMessage("1 answer found.");
	    }
	    else
	    {
		parentWin.displayMessage(qr.numResults() + " answers found.");
	    }

	    resultList = new MultiColumn(qr.getNames(), this);
	    
            add("North", outputIndicators);
            add("Center", resultList);
            validate();

            putAnswersInMultiList(qr);
        }

	endTime = new Date();
	long duration = endTime.getTime() - startTime.getTime();
	System.out.println("User time: " + duration + " ms");
	pn.getTime();
	System.out.println();

	return qr;
    }
    */

    /* Called on a single click on a list item.
       Rows are numbered starting with 0.
       If the click is on an unselected row, then it shows the
       corresponding bindings in the graph. If the row was previously
       selected then it turns off the bindings in the graph.
    */
    public void singleClick (MultiColumn list, int row, int col)
    {
        if (row >= 0)
	{
	   if (row != selectedRow)
           {
	     String[]  binds = list.getItem(row);
	     da.setGraphVariableBindings (binds, list.title);
	     selectedRow = row;
	   }
	   else 
           {   
	     da.unsetGraphVariableBindings();
	     selectedRow = -1;
	   }
	}
    }

    /* Called upon double-clicking on a list item.
       If running as an applet in a web browser and if the cell clicked
       on is a URL, loads the web page into the browser
    */
    public void doubleClick(MultiColumn list, int row, int col)
    {
      URL url = null;

      // only works for applets add when a cell in the MultiColumn is selected
      if (context != null && row >= 0 && col >= 0) { 
	/* retrieve bindings  */
	String[] binds = list.getItem(row);
	if (binds[col].startsWith("http://")) {         // probably a URL
	  try {
	    url = new URL(binds[col]);
	    // System.out.println("Trying to show " + url);
	    // display document in a new window named "SHOE_Doc"
	    context.showDocument(url, "SHOE_Doc");
	  } catch (MalformedURLException ex) {
	    System.out.println(url + " not valid"); 
	  }
	}
      }
    }  

/*  not used currently
    private  void initAnswersForMultiList()
    {
	totalNumAnswers = pn.getNumTotalAnswersToQuery();
        numCols = pn.getNumVariables();

	if (prevCB.isEnabled())
	    prevCB.disable();

	if (totalNumAnswers > blockSize)
	    nextCB.enable();
	else if (nextCB.isEnabled())
	    nextCB.disable();

	// call for first set of answers for the user.
	firstMultiListItem = 1;
        lastMultiListItem = Math.min (blockSize, totalNumAnswers);
	setTextFields();
	putAnswersInMultiList();
        validate();
    }
//
//  not used currently, either:

    //Must be called sometime after pn.submitQuery
    //and after firstMultiListItem and lastMultiListItem
    //are correctly set.
    //
    private  void putAnswersInMultiList()
    {
	boolean  haveError = false;

	// If we do not call 'MultiColumn.clear()', answers just
	// accumulate.  Of course that would require different handling
	// than we do here, since when a user would go back and forward
	// multiple times with the current way of doing things, we'd
	// just accumulate more and more answers (same ones appending).
	//
	resultList.clear();

        pn.getAnswersFromServer (firstMultiListItem,
                           lastMultiListItem * numCols);

        String[] answers = pn.getAnswers();
// TEST
//	for (int k = 0; k < answers.length; k++)
//	    System.out.println (answers[k]);
//
	int numRows = (lastMultiListItem - firstMultiListItem) + 1;
	for (int i = 0; i < numRows; i++)
	{
	    String[]  oneRow = new String[numCols];
	    for (int j = 0; j < numCols; j++)
	    {
		try
		{
		    oneRow[j] = answers[(i*numCols)+j];
		}
		catch (ArrayIndexOutOfBoundsException  exc)
		{
		    System.out.println ("Nary Parka Error: " +
		                       exc.getLocalizedMessage());
		    exc.printStackTrace();
		    oneRow[j] = new String("Error");
		    haveError = true;
		}
            }
	    resultList.addItem (oneRow);
	    if (haveError)
	        break;   // do not continue to spew out error messages
	}
	validate();
	parentWin.validate();
    }


  /* This function takes a string intended as an argument and escapes those
     characters that have special meaning in a Parka Assertion or query.
     In particular, the hash ('#'), backslash ('\') and parentheses ('(', ')')
     are escaped. Also, newlines are converted to spaces so that the 
     query string can be safely used in the file format */
  String escapeSpecialChars(String inString) {

    StringBuffer outString;
    int lastEsc = 0;
    int i;
    char c;

    outString = new StringBuffer(inString.length());
    for (i=0; i<inString.length(); i++) {
      c = inString.charAt(i);
      if (c == '#' || c =='\\' || c=='(' || c==')') {
	outString.append(inString.substring(lastEsc, i));
	outString.append("\\");         // insert a backslash before c
	lastEsc = i;
      }
      if (c == '\n') {
	outString.append(" ");
	lastEsc = i + 1;
      }
    }
    outString.append(inString.substring(lastEsc));
    return outString.toString();
  }
	    
}
