/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
package parka;

import java.awt.*;

/**********************************************************************
  toolbar
  **********************************************************************/

public class ToolBar extends java.awt.Panel
{
    private QueryWin  parent;
    private Tools     tools;
    private TextField text;
    private Checkbox  isvariable;
    private DrawArea  drawArea;

    private static int HORIZBORDERAREA = 4;
    private static int VERTBORDERAREA  = 4;

    /*******
     Methods
     *******/

    public ToolBar (QueryWin MyWin)
    {
        parent = MyWin;
        drawArea = null;

        // Create toolbar components
        tools = new Tools (MyWin);
        text = new TextField();
        setTextDisable();
        isvariable = new Checkbox("Variable");
        isvariable.disable();

	GridBagLayout gb = new GridBagLayout();
        setLayout (gb);

	GridBagConstraints constraints = new GridBagConstraints();

	setGridbagConstraints (constraints, 0, 1, 0.0);
        gb.setConstraints (tools, constraints);
        add (tools);

	setGridbagConstraints (constraints, 1, 1, 1.0);
	gb.setConstraints (text, constraints);
	add (text);

	setGridbagConstraints (constraints, 2, 1, 0.0);
        gb.setConstraints (isvariable, constraints);
	add (isvariable);
    }

    private void setGridbagConstraints (GridBagConstraints gbc,
                 int gridxPos, int gridwid, double xWt)
    {
        gbc.gridx = gridxPos;
        gbc.gridy = 0;
        gbc.gridwidth = gridwid;
        gbc.gridheight = 1;
        gbc.weightx = xWt;
        gbc.weighty = 0;
        gbc.fill = GridBagConstraints.HORIZONTAL;
        gbc.anchor = GridBagConstraints.CENTER;
        gbc.insets.left = gbc.insets.right = HORIZBORDERAREA;
        gbc.insets.top = gbc.insets.bottom = VERTBORDERAREA;
    }

    void  setDrawArea (DrawArea  da)
    {
	drawArea = da;
    }

    void  resetCursor()
    {
	tools.resetCursor();
    }

    int  getCurrentTool()
    {
	return  tools.getCurrentTool();
    }

    void  setCurrentTool (int newtool)
    {
	if (!tools.isLocked())
	    tools.changeTool (newtool);
    }

    // Modify string if it contains any invalid characters
    public String validateStr (String val)
    {
        return val.replace('#', '_');
    }

    //has to have selected all of text in TextField to return true.
    boolean textIsSelected()
    {
	return  ((text.getSelectionStart() == 0) &&
                  ((text.getSelectionEnd() +1) == (text.getText()).length()) );
    }

    String getTextContents()
    {
	return   (text.getText());
    }

    public boolean mouseEnter (Event evt, int x, int y)
    {
        resetCursor();
            //make sure correct cursor is drawn for
            //current tool
        return true;
    }

    // event handler for return key (for text field dump)
    public boolean action (Event evt, Object arg)
    {
        if (evt.target instanceof TextField)
        {
            // Test to see if the value is valid
	    //String val = this.validateStr(text.getText());
	  /* removed this so that '#' can be used in queries, they will
	     be escaped before being sent to the server */

            String val = text.getText();
            text.setText(val);

            if (val.length() > 0) 
            {
		if (drawArea != null)
		{
                    Node.changeText (val, drawArea);
                    drawArea.requestFocus();
		}
            }
            return true;
        }
        // event handler for checkbox
        else if (evt.target instanceof Checkbox)
        {
            if (evt.id == Event.ACTION_EVENT)
                drawArea.toggleVarStatus (isvariable.getState());
            return true;
        }
        
        return false;
    }

    void activateTextField()
    {
	// prepare for changing active Node's name
        selectText();   //highlights text
	//send focus to TextField
        requestTextFocus();
    }

    void deactivateNodeDisplays()
    {
        setText ("");
        setTextDisable();
        setVariableState (false);
        setVariableDisable();
    }

    void activateNodeDisplays (Node n)
    {
        String   nodeName = n.getObjName();
	boolean  varState = n.getIsVariable();

        setText (nodeName);
        setTextEnable();
        setVariableState (varState);
        setVariableEnable();
    }

    void setVariableState (boolean  newState)
    {
	isvariable.setState (newState);
    }

    void setVariableEnable()
    {
	isvariable.enable();
    }

    void setVariableDisable()
    {
	isvariable.disable();
    }

    private void setTextDisable()
    {
	text.setEditable(false);
    }

    private void setTextEnable()
    {
	text.setEditable(true);
    }

    void setText (String str)
    {
	text.setText (str);
    }

    void requestTextFocus()
    {
	text.requestFocus();
    }

    void selectText()
    {
	text.selectAll();
    }

    public void displayMessage(String msg)
    {
        // Displays a message in the text area
        text.setText(msg);
    }
}
