/* Copyright (C) 1996-1999 PLUS Group - University of Maryland

   This software is provided under the terms of the GNU General Public
   License Version 2, June 1991. See the file License.html for details.
   */
/**********************************************************************
  Tools class
  **********************************************************************/
package parka;

import java.awt.*;
import java.awt.image.*;
import java.io.IOException;
import java.net.*;

class Tools extends java.awt.Canvas
{
    private QueryWin     parent;
    private URL          base;
    private Image        images[];

    private  int      tool;
    private  boolean  locked;

    // these are numbered in the order they are displayed
    //
    final static int MOVE   = 0;
                 //without mouse movement, merely makes the active ob.
    final static int NODE   = 1;
    final static int LINK   = 2;
    final static int EDGE   = 3;

    final static int NUMTOOLS = 4;
    final static int NUMIMAGES_PER_TOOL = 3;

    private  final static int BUTTONLENGTH = 24;  // button sizes in pixels

    // three images for each button.
    // one for unselected; one for selected-for-one-operation-only; and
    // one for selected-and-locked (keep selected tool until another is
    // explicitly chosen by the user)
    //
    final static String toolpaths[] =
    {
        "cursor.gif",
        "cursor2.gif",
        "cursor3.gif",
        "variable.gif",
        "variable2.gif",
        "variable3.gif",
        "relation.gif",
        "relation2.gif",
        "relation3.gif",
        "line.gif",
        "line2.gif",
        "line3.gif"
    };

    final static String  GIF_DIRECTORY = "gifs/";

    Tools (QueryWin MyWin)
    {
        parent = MyWin;
	locked = false;
        tool = MOVE; // init tool
	images = new Image [NUMTOOLS * NUMIMAGES_PER_TOOL];

        // if this is an applet:
        if (ApplType.getIsApplet())
        {
            base = ((Parka)parent.parent).getCodeBase();
	    URL   eachtoolLoc;

	    for (int x = 0; x < NUMTOOLS * NUMIMAGES_PER_TOOL; x++)
	    {
		try   {
		    eachtoolLoc = new URL (base, GIF_DIRECTORY + toolpaths[x]);
		    images[x] = ((Parka)parent.parent).getImage (
                                     eachtoolLoc);
/* TEST
                    System.out.println("in Tools constructor:");
                    System.out.println("  Next graphic " + x +
                                " for tool is: " + eachtoolLoc);
*/
		}
		catch (MalformedURLException mu_e)
                {
		    images[x]= null;
		    System.out.println ("Could not load one of Tool Images");
		    System.out.println (mu_e.getMessage());
                }
	    }
        }
        // if this is an application:
        else  if (ApplType.getIsApplication())
        {
            ImageProducer  ip;

	    for (int x = 0; x < NUMTOOLS * NUMIMAGES_PER_TOOL; x++)
            {
                base = ((ParkaApp)parent.parent).get_url_for_icons (
                                        toolpaths[x]);
		
                try  {
                    ip = (ImageProducer)base.getContent();
                }
                catch (IOException io_e)    {
                    System.out.println (
                        "Failed to get icons for Query Creation Actions");
                    ip = null;
                }
                if (ip != null)
                    images[x] = createImage (ip);
	    }
        }
        else
        {
            for (int x = 0; x < NUMTOOLS * NUMIMAGES_PER_TOOL; x++)
		images[x] = null;
        }
        resize (BUTTONLENGTH * NUMTOOLS, BUTTONLENGTH); //width, height
    }

    boolean  isLocked()
    {
	return  locked;
    }

    void changeTool (int newTool)
    {
        tool = newTool;
        switch (tool)
        {
            case Tools.MOVE:
                    parent.setCursor(Frame.DEFAULT_CURSOR);
                    break;
            case Tools.NODE:
                    parent.setCursor(Frame.HAND_CURSOR);
                    break;
            case Tools.LINK:
                    parent.setCursor(Frame.MOVE_CURSOR);
                    break;
            case Tools.EDGE:
                    parent.setCursor(Frame.CROSSHAIR_CURSOR);
	}
        repaint();
    }

    void resetCursor()
    {
        switch (tool)
        {
            case Tools.MOVE:
                    parent.setCursor(Frame.DEFAULT_CURSOR);
                    break;
            case Tools.NODE:
                    parent.setCursor(Frame.HAND_CURSOR);
                    break;
            case Tools.LINK:
                    parent.setCursor(Frame.MOVE_CURSOR);
                    break;
            case Tools.EDGE:
                    parent.setCursor(Frame.CROSSHAIR_CURSOR);
	}
        repaint();
    }

    public int getCurrentTool()
    {
	return  tool;
    }

    // override method in class Canvas
    public void paint (Graphics g)
    {
        Rectangle b = bounds();

        // get width margin on one side
        if (b.width < BUTTONLENGTH * NUMTOOLS)
            b.width = 0;
        else b.width = (b.width - BUTTONLENGTH * NUMTOOLS)/2;

        // get height margin on one side
        if (b.height < BUTTONLENGTH)
            b.height = 0;
        else b.height = (b.height - BUTTONLENGTH)/2;

        if (g == null)
	    return;

        super.paint(g);
        g.setPaintMode();

	for (int x = 0; x < NUMTOOLS * NUMIMAGES_PER_TOOL; x++)
	    prepareImage (images[x], this);

	for (int x = 0; x < NUMTOOLS; x++)
        {
	    // x will == MOVE, NODE, LINK, or EDGE:
	    if ((tool == x) && locked)
		g.drawImage (images[x*3+2], BUTTONLENGTH*x + b.width,
                                 b.height, this);
	    else   if (tool == x)
		g.drawImage (images[x*3+1], BUTTONLENGTH*x + b.width,
                                 b.height, this);
	    else  //not selected at all:
		g.drawImage (images[x*3], BUTTONLENGTH*x + b.width,
                                 b.height, this);
        }
    }

    public boolean mouseDown (Event evt, int x, int y)
    {
        // Note: clickCount only ever non-zero for mouseDown events
	if (evt.clickCount == 1)
	    locked = false;
	else  if (evt.clickCount > 1)
	    locked = true;

        Rectangle b = bounds();

        if (b.width < BUTTONLENGTH * NUMTOOLS)
            b.width = 0;
        else b.width = (b.width - BUTTONLENGTH * NUMTOOLS)/2;

        if (b.height < BUTTONLENGTH)
            b.height = 0;
        else b.height = (b.height - BUTTONLENGTH)/2;

        for (int a = 0; a < NUMTOOLS; a++)
        {
            // hit tests
            if ((x >= a * BUTTONLENGTH) && (x < (a+1) * BUTTONLENGTH))
            {
                changeTool (a);
                paint (this.getGraphics());
                return   true;
            }
        }
        return  false;
    }
}
